
import fish_feats.Utils as ut
import fish_feats.FishWidgets as fwid
from fish_feats.NapaMix import Separation
import numpy as np
from qtpy.QtWidgets import QVBoxLayout, QWidget 
from napari.utils.translations import trans

class MeasureNuclei( QWidget ):
    """ Measure intensity inside segmented nuclei """

    def __init__( self, viewer, mig, cfg ):
        """ Interface to measure nuclear intensities """

        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
        self.results = None
        
        super().__init__()
        layout = QVBoxLayout()

        ## choose channel to measure
        chan_line, self.channel_spin = fwid.spinner_line( "Channel: ", 0, self.mig.nbchannels, 1, 0, changefunc=self.show_layer, descr="Choose which channel to measure" )
        layout.addLayout( chan_line )

        ## Go, measure
        measure_btn = fwid.add_button( "Measure", btn_func=self.go_measure, descr="Measure intensity in nuclei in selected channel" )
        layout.addWidget( measure_btn )

        ## show or save the results
        table_btn = fwid.add_button( "Show measure table", btn_func=self.show_results, descr="Open a window with the results table with all nuclei measured" )
        table_save_btn = fwid.add_button( "Save all nuclei measures", self.save_table, descr="Save a table file with the measure of ALL nuclei (even not associated)" )
        double_line = fwid.double_widget( table_btn, table_save_btn )
        layout.addLayout( double_line )

        
        ## Done, save and quit
        done_btn = fwid.add_button( "Save and stop", btn_func=self.done, descr="Add measurements to the Results file and quit the option", color=ut.get_color("done") )
        layout.addWidget( done_btn )

        self.setLayout( layout )

        text = "Measure intensity inside segmented nuclei \n"
        ut.showOverlayText( self.viewer, text )
        print("******** Measure nuclear intensity ******")

    def show_layer( self ):
        """ Change currently visible layer based on the current channel selection """
        for lay in self.viewer.layers:
            lay.visible = False 
        chan = self.channel_spin.value()
        if "originalChannel"+str(chan) in self.viewer.layers:
            self.viewer.layers["originalChannel"+str(chan)].visible = True

    def go_measure( self ):
        """ Do the intensity measurement """
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=None, descr="Measuring nuclei..." ) 
        chan = self.channel_spin.value()
        self.results = self.mig.measure_nuclear_intensity( chan )
        ut.show_info("Measure done")
        ut.close_progress( self.viewer, pbar )
        ut.show_duration( start_time, "Nuclear intensities measured in " )

    def save_table( self ):
        """ Save the results of all nuclei measures """
        if self.results is None:
            ut.show_warning("No measurement done yet")
            return
        chan = self.channel_spin.value()
        outname = self.mig.build_filename( "_allnuclei_C"+str(chan)+".csv" )
        ut.write_dict( outname, self.results )
        ut.show_info("Results saved in file "+str(outname))


    def show_results( self ):
        """ Show the table of resutls of all nuclei """
        if self.results is None:
            ut.show_warning("No measurement done yet")
            return
        ut.show_table( self.results )

    def done( self ):
        """ Save results and quit option """
        self.mig.save_results()
        ut.remove_widget( self.viewer, "Measure nuclei" )


#### Nuclei widget class
class NucleiWidget(QWidget):
    """ Widget to get nuclei options """

    def __init__( self, ffeats ):
        super().__init__()
        self.mig = ffeats.mig
        self.viewer = ffeats.viewer
        self.cfg = ffeats.cfg
        self.ffeats = ffeats

        ## Load default/configuration parameters
        self.load_parameters()
        self.nuclei_filename = self.mig.nuclei_filename(ifexist=True)
        if self.nuclei_filename is None:
            self.nuclei_filename = self.mig.resdir

        layout = QVBoxLayout()
        ## Method choice
        method_line, self.method_choice = fwid.list_line( "Method: ", descr="Choose method to get nuclei: load or segment", func=self.show_parameters_nuclei )
        layout.addLayout( method_line )

        ## Stardist parameters
        self.stardist_group, stardist_layout = fwid.group_layout( "Stardist parameters", descr="", color = ut.get_color("group1") )
        prob_line, self.stardist_probability_threshold = fwid.value_line( "Probability threshold: ", self.paras["Stardist_probability_threshold"], descr="Probability threshold to keep pixel in a nucleus (the lower the more/larger nuclei)" )
        stardist_layout.addLayout( prob_line )
        overlap_line, self.stardist_nuclei_overlap = fwid.value_line( "Nuclei overlap: ", self.paras["Stardist_nuclei_overlap"], descr="Overlap threshold to consider two nuclei as overlapping (the lower the more/larger nuclei)" )
        stardist_layout.addLayout( overlap_line )
        ##2D->3D association
        association_line, self.stardist_association_method = fwid.list_line( "3D reconstruction method: ", descr="Method to associate 2D Stardist nuclei into 3D nuclei in consecutive slices" )
        for method in ["Munkres", "Overlap"]:
            self.stardist_association_method.addItem( method )
        ind = self.stardist_association_method.findText( "Overlap" )
        if ind >= 0:
            self.stardist_association_method.setCurrentIndex( ind )
        stardist_layout.addLayout( association_line )
        distance_line, self.stardist_association_distance_limit_micron = fwid.value_line( "Max distance to associate (microns): ", self.paras["Stardist_association_distance_limit_micron"], descr="Distance limit to associate two 2D nuclei across consecutive slices" )
        stardist_layout.addLayout( distance_line )
        threshold_line, self.stardist_threshold_overlap = fwid.value_line( "Threshold overlap: ", self.paras["Stardist_threshold_overlap"], descr="Threshold to consider two nuclei as overlapping (high => small divided nuclei)" )
        stardist_layout.addLayout( threshold_line )
        self.stardist_group.setLayout( stardist_layout )
        layout.addWidget( self.stardist_group )
        
        ## CellPose3D parameters
        self.cp_detthreshold_div = 10
        self.cp_stitchthreshold_div = 100
        self.cellpose_group, cellpose_layout = fwid.group_layout( "CellPose3D parameters", descr="", color = ut.get_color("group2") )
        cp_diameter_line, self.cellpose_cell_diameter = fwid.value_line( "Cell diameter (pixels): ", self.paras["Cellpose_cell_diameter"], descr="Average diameter of the cells (in pixels)" )
        cellpose_layout.addLayout( cp_diameter_line )
        cp_detection_line, self.cellpose_detection_threshold = fwid.slider_line( "Detection threshold: ", minval=-6.0, maxval=6.0, step=0.2, value=self.paras["Cellpose_detection_threshold"], show_value=True, slidefunc=None, descr="Detection threshold for CellPose (the lower the more/larger nuclei)", div=self.cp_detthreshold_div )
        cellpose_layout.addLayout( cp_detection_line )
        self.cellpose_resample = fwid.add_check( "Resample: ",  checked=self.paras["Cellpose_resample"],
        check_func=None,
        descr="Resample CP output before to compute the cell contours (slower but more precise)" )
        cellpose_layout.addWidget( self.cellpose_resample )
        self.cellpose_in3D = fwid.add_check( "In 3D: ",  checked=self.paras["Cellpose_in3D"],
        check_func=None,
        descr="Use CellPose in 3D mode (XY, YZ, XZ), otherwise in 2D+stitching" )
        cellpose_layout.addWidget( self.cellpose_in3D )
        cp_stitch_line, self.cellpose_stitch_threshold = fwid.slider_line( "Stitching threshold: ", minval=0, maxval=1, step=0.01, value=self.paras["Cellpose_stitch_threshold"], show_value=True, slidefunc=None, descr="Threshold to stitch 2D CellPose nuclei into 3D nuclei (the lower the larger nuclei)", div=self.cp_stitchthreshold_div )
        cellpose_layout.addLayout( cp_stitch_line )

        ## dask line
        self.cellpose_dask = fwid.add_check( "Use dask with chunk size:", self.paras["Cellpose_dask"], None, descr="Use Cellpose-Dask version to compute faster or without memory issues" )
        self.cellpose_chunk_size = fwid.add_value( self.paras["Cellpose_chunk_size"], descr="Chunk size when using CellPose-dask" )
        dask_line = fwid.double_widget( self.cellpose_dask, self.cellpose_chunk_size )
        cellpose_layout.addLayout( dask_line )

        self.cellpose_group.setLayout( cellpose_layout )
        layout.addWidget( self.cellpose_group )

        ## Load segmented file parameters
        self.load_file_group, load_file_layout = fwid.group_layout( "Load segmented file", descr="Load a file with segmented nuclei", color = ut.get_color("group3") )
        filename_line, choose_filename, self.filename_label = fwid.label_button( "Choose file", self.get_nuclei_filename, label=self.nuclei_filename, descr="Choose a file with segmented nuclei", color=None )
        load_file_layout.addLayout( filename_line )
        self.load_file_group.setLayout( load_file_layout )
        layout.addWidget( self.load_file_group )

        go_line, go_btn = fwid.line_button_help( "Get nuclei",
            self.get_nuclei, 
            "Segment nuclei with CellPose3D/Stardist, or load a segmented file",
            help_link="Get-nuclei", 
            color = ut.get_color("go"),
            )
        layout.addLayout( go_line )
        for method in ["CellPose3D", "Stardist", "Load segmented file"]:
            self.method_choice.addItem( method )
        ind = self.method_choice.findText( self.defmeth )
        if ind >= 0:
            self.method_choice.setCurrentIndex( ind )
        self.show_parameters_nuclei()
        self.setLayout(layout)

    def get_nuclei_filename( self ):
        """ Open a file dialog to choose a file with segmented nuclei """
        filename = fwid.file_dialog( "Choose segmented nuclei file", "*.tif", directory=self.mig.resdir )
        if filename is not None:
            self.nuclei_filename = filename
            self.filename_label.setText(filename)

    def load_parameters( self ):
        """ Load default parameters and update from the configuration file """
        ## init/load parameters
        self.defmeth = "Stardist"
        self.paras = {}
        self.paras["Stardist_probability_threshold"] = 0.5
        self.paras["Stardist_nuclei_overlap"] = 0.1
        self.paras["Stardist_association_distance_limit_micron"] = 0.3
        self.paras["Stardist_threshold_overlap"] = 0.2
        self.paras["Cellpose_cell_diameter"] = 8
        self.paras["Cellpose_detection_threshold"] =0.0
        self.paras["Cellpose_resample"] = True 
        self.paras["Cellpose_in3D"] = True
        self.paras["Cellpose_dask"] = False 
        self.paras["Cellpose_stitch_threshold"] = 0.25
        self.paras["Cellpose_chunk_size"] = 1000
        load_paras = self.cfg.read_parameter_set("NucleiSeg")
        if load_paras is not None:
            float_paras = ["Cellpose_cell_diameter", 
                           "Cellpose_chunk_size",
                           "Cellpose_detection_threshold", "Cellpose_stitch_threshold", "Stardist_probability_threshold", "Stardist_threshold_overlap", "Stardist_nuclei_overlap", "Stardist_association_distance_limit_micron"]
            
            for cpara in float_paras:
                if cpara in load_paras:
                    self.paras[cpara] = float(load_paras[cpara])
            if "Cellpose_resample" in load_paras:
                self.paras["Cellpose_resample"] = load_paras["Cellpose_resample"].strip() == "True" 
            if "Cellpose_in3D" in load_paras:
                self.paras["Cellpose_in3D"] = load_paras["Cellpose_in3D"].strip() == "True" 
            if "Cellpose_dask" in load_paras:
                self.paras["Cellpose_dask"] = load_paras["Cellpose_dask"].strip() == "True" 
            if "method" in load_paras:
                self.paras["method"] = load_paras["method"]
                self.defmeth = load_paras["method"]
    
        if self.mig.nuclei_filename(ifexist=True) != "":
            self.defmeth = "Load segmented file"

    def show_parameters_nuclei( self ):
        """ Controls the visibility of paratemeters according to the method """
        self.stardist_group.setVisible( self.method_choice.currentText() == "Stardist" )
        self.cellpose_group.setVisible( self.method_choice.currentText() == "CellPose3D" )
        self.load_file_group.setVisible( self.method_choice.currentText() == "Load segmented file" )

    def nuclei_stardist( self ):
        """ Do segmentation with Stardist """
        self.viewer.text_overlay.text = trans._( """ Doing segmentation of cell nuclei with Stardist... \n Prepare nuclei image """ )
        self.cfg.addText("Segment nuclei with Stardist2D+3D association")
        self.cfg.addTextParameter( "Stardist", "probability_threshold", self.stardist_probability_threshold.text() )
        self.cfg.addTextParameter( "Stardist", "nuclei_overlap", self.stardist_nuclei_overlap.text() )
        self.cfg.addTextParameter( "Stardist", "association_distance_limit_micron", self.stardist_association_distance_limit_micron.text() )
        self.cfg.addTextParameter( "Stardist", "threshold_overlap", self.stardist_threshold_overlap.text() )
        self.mig.prepare_segmentation_nuclei()
        self.viewer.text_overlay.text = trans._( """ Doing segmentation of cell nuclei with Stardist... \n Do Stardist2D+assocation3D """ )
        self.go_segnuclei_stardist()
    
    def go_segnuclei_stardist( self ):
        """ Launch segmentation with StarDist """
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=None, descr="Segmenting with Stardist.." )
        self.mig.do_segmentation_stardist( float(self.stardist_probability_threshold.text() ), 
            float(self.stardist_nuclei_overlap.text()),
            self.stardist_association_method.currentText(), 
            float(self.stardist_association_distance_limit_micron.text()),
            float(self.stardist_threshold_overlap.text()) )
        if self.mig.nucmask is None:
            ut.close_progress( self.viewer, pbar )
            return
        self.end_segmentation_nuclei( start_time, pbar )

    
    def nuclei_cellpose( self ):
        """ Do segmentation with CellPose """
        self.viewer.text_overlay.text = trans._( """ Doing segmentation of cell nuclei with CellPose3D... \n Prepare nuclei image """ )
        self.mig.prepare_segmentation_nuclei()
        self.viewer.text_overlay.text = trans._( """ Doing segmentation of cell nuclei with CellPose3D... \n Doing Cellpose3D... """ )
        ## add parameters to the configuration file
        self.cfg.addText("Segment nuclei with CellPose 3D")
        self.cfg.addTextParameter( "CellPose", "cell_diameter", self.cellpose_cell_diameter.text() )
        self.cfg.addTextParameter( "CellPose", "detection_threshold", self.cellpose_detection_threshold.value()*1.0/self.cp_detthreshold_div )
        self.cfg.addTextParameter( "CellPose", "resample", self.cellpose_resample.isChecked())
        self.cfg.addTextParameter( "CellPose", "in3D", self.cellpose_in3D.isChecked() )
        self.cfg.addTextParameter( "CellPose", "dask", self.cellpose_dask.isChecked() )
        self.cfg.addTextParameter( "CellPose", "stitch_threshold", self.cellpose_stitch_threshold.value()*1.0/ self.cp_stitchthreshold_div )
        self.cfg.addTextParameter( "CellPose", "chunk_size", int(float(self.cellpose_chunk_size.text())) )
        self.go_segnuclei_cellpose()

    def go_segnuclei_cellpose( self):
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=None, descr="Segmenting with CellPose.." )
        self.mig.do_segmentation_cellpose( float(self.cellpose_cell_diameter.text()),
        float(self.cellpose_detection_threshold.value())/self.cp_detthreshold_div,
        self.cellpose_resample.isChecked(),
        self.cellpose_in3D.isChecked(),
        float(self.cellpose_stitch_threshold.value())/ self.cp_stitchthreshold_div,
         dask=self.cellpose_dask.isChecked(),
          chunk = int(float(self.cellpose_chunk_size.text())) )
        self.end_segmentation_nuclei( start_time, pbar )

    def load_segmentationfile_nuclei( self ):
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=None, descr="Loading nuclei file.." )
        self.mig.load_segmentation_nuclei( self.nuclei_filename )
        self.end_segmentation_nuclei( start_time, pbar )

    def get_nuclei( self ):
        """ Interface to choose parameters of 3d nuclei segmentation """
        ut.remove_layer( self.viewer, "segmentedNuclei" )
        self.cfg.addGroupParameter("NucleiSeg")
        self.cfg.addParameter( "NucleiSeg", "method", self.method_choice.currentText() )
        self.cfg.addParameter( "NucleiSeg", "Stardist_probability_threshold", self.stardist_probability_threshold.text() )
        self.cfg.addParameter( "NucleiSeg", "Stardist_nuclei_overlap", self.stardist_nuclei_overlap.text() )
        self.cfg.addParameter( "NucleiSeg", "Stardist_association_distance_limit_micron", 
            self.stardist_association_distance_limit_micron.text() )
        self.cfg.addParameter( "NucleiSeg", "Stardist_threshold_overlap", self.stardist_threshold_overlap.text() )
        self.cfg.addParameter( "NucleiSeg", "Cellpose_cell_diameter", self.cellpose_cell_diameter.text() )
        self.cfg.addParameter( "NucleiSeg", "Cellpose_detection_threshold", self.cellpose_detection_threshold.value()*1.0/self.cp_detthreshold_div )
        self.cfg.addParameter( "NucleiSeg", "Cellpose_resample", self.cellpose_resample.isChecked())
        self.cfg.addParameter( "NucleiSeg", "Cellpose_dask", self.cellpose_dask.isChecked())
        self.cfg.addParameter("NucleiSeg", "Cellpose_in3D", self.cellpose_in3D.isChecked() )
        self.cfg.addParameter( "NucleiSeg", "Cellpose_stitch_threshold", self.cellpose_stitch_threshold.value()*1.0/ self.cp_stitchthreshold_div )
        self.cfg.write_parameterfile()
        ut.removeOverlayText(self.viewer)

        if self.method_choice.currentText() == "Load segmented file":
            ut.show_info("Load nuclei from file "+str(self.nuclei_filename))
            self.load_segmentationfile_nuclei()
        else:
            if self.mig.should_separate():
                ut.show_info("Junctions and nuclei staining in the same channel, separate them first")
                separation = Separation( self.viewer, self.mig, self.cfg ) 
                self.viewer.window.add_dock_widget(separation, name="Separate")
            else:
                if self.method_choice.currentText() == "CellPose3D":
                    ut.showOverlayText( self.viewer, "Doing segmentation of cell nuclei...")
                    self.nuclei_cellpose()
                if self.method_choice.currentText() == "Stardist":
                    ut.showOverlayText( self.viewer, "Doing segmentation of cell nuclei...")
                    self.nuclei_stardist()


    def end_segmentation_nuclei( self, start_time, pbar ):
        """ Finished segmentation, go to manual correction step """
        ut.show_duration( start_time, "Nuclei segmentation/loading took " )
        ut.close_progress( self.viewer, pbar )
        ut.removeOverlayText(self.viewer)
        ut.remove_widget(self.viewer, "Get nuclei")
        self.correction_nuclei()

    def correction_nuclei( self ):
        """ Allows manual correction of the segmentation """
        ut.show_info("Correct nuclei segmentation if necessary")
        #self.viewer.add_image( self.mig.nucstain, name="nucleiStaining", scale=( self.mig.scaleZ,self.mig.scaleXY,self.mig.scaleXY), blending="additive" )
        maskview = self.viewer.add_labels( self.mig.nucmask, blending='additive', scale=( self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY ), name="segmentedNuclei" )
        maskview.n_edit_dimensions = 3
        maskview.contour = 0
        filter_nuclei = FilterNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget( filter_nuclei, name="Filtering" )
        self.ffeats.showCellsWidget("segmentedNuclei", shapeName="NucleiName", dim=3)
        finish_nuc = FinishNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget( finish_nuc, name="End nuclei" )

class FilterNuclei( QWidget ):
    """ Filter to remove/correct nuclei segmentation """

    def __init__( self, viewer, mig, cfg ):
        """ Interface to filter nuclei options """
        
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
    
        paras = self.cfg.read_parameter_set("FilterNuclei")
        minvol = 50.0
        keepnz = 5
        rmsmall = True
        if paras is not None:
            if "minimum_volume" in paras:
                minvol = float(paras["minimum_volume"])
            if "keep_ifatleast_z" in paras:
                keepnz = int(paras["keep_ifatleast_z"])
            if "remove_small_nuclei" in paras:
                rmsmall = paras["remove_small_nuclei"].strip() == True
        
        super().__init__()
        layout = QVBoxLayout()

        ## remove small nuclei
        self.remove_small = fwid.add_check( "Remove nuclei smaller than ", rmsmall, None, descr="If checked, nuclei smaller than given threshold will be removed" )
        self.size_threshold = fwid.add_value( minvol, descr="Minimum volume (in pixels) to keep a nucleus" )
        small_line = fwid.double_widget( self.remove_small, self.size_threshold )
        layout.addLayout( small_line )

        ## keep if mim nb of Z
        nbline, self.min_nb_z = fwid.value_line( "Min nb of Z in nucleus", keepnz, descr="Keep only nuclei that are present in more Z slices than the minimal number" )
        layout.addLayout( nbline )

        ## Go
        go_btn = fwid.add_button( "Update nuclei", self.go_filter, descr="Filter nuclei based on current parameters" )
        layout.addWidget( go_btn )
        self.setLayout( layout )

    def go_filter( self ):
        """ Apply the filtering """
        if not self.remove_small.isChecked():
            mvol = -1
        else:
            mvol = int(float(self.size_threshold.text()))

        print("--- Filtering nuclei ---")
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=None, descr="Filtering nuclei..." )
        self.mig.filterNuclei( mvol, (float(self.min_nb_z.text())) )
        ut.close_progress( self.viewer, pbar )
        ut.show_duration( start_time, "Nuclei filtered took ")

        self.cfg.addGroupParameter("FilterNuclei")
        self.cfg.addParameter("FilterNuclei", "minimum_volume", (float(self.size_threshold.text())) )
        self.cfg.addParameter( "FilterNuclei", "keep_ifatleast_z", int(float(self.min_nb_z.text())) )
        self.cfg.addParameter("FilterNuclei", "remove_small_nuclei", self.remove_small.isChecked() )
        self.cfg.write_parameterfile()


class FinishNuclei( QWidget ):
    """ Filter to remove/correct nuclei segmentation """

    def __init__( self, viewer, mig, cfg ):
        """ Interface to filter nuclei options """
        
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
    
        super().__init__()
        layout = QVBoxLayout()

        ## save button
        save_btn = fwid.add_button( "Save nuclei", self.save_nuclei, descr="Save the current nuclei segmentation", color=ut.get_color("save") )
        ## done btn
        done_btn = fwid.add_button( "Nuclei done", self.nuclei_done, descr="Finish this step, close everything related", color=ut.get_color("done") )
        line = fwid.double_widget( save_btn, done_btn )
        layout.addLayout( line )

        self.setLayout( layout )

    def save_nuclei( self ):
        """ Save the current segmentation of nuclei to file """
        if 'segmentedNuclei' in self.viewer.layers:
            outname = self.mig.nuclei_filename(ifexist=False)
            self.mig.save_image( self.viewer.layers["segmentedNuclei"].data, outname, hasZ=True )

    def nuclei_done( self ):
        """ Finish the nuclei editing step, close everything """
        ut.remove_layer( self.viewer,"nucleiStaining" )
        ut.remove_layer( self.viewer,"segmentedNuclei" )
        ut.remove_layer( self.viewer,"NucleiName" )
        ut.removeOverlayText( self.viewer )
        #addText("")
        ut.remove_widget(self.viewer, "Filtering")
        ut.remove_widget(self.viewer, "CellPose3D")
        ut.remove_widget(self.viewer, "Stardist")
        ut.remove_widget(self.viewer, "Edit labels")
        ut.remove_widget(self.viewer, "End nuclei")



class PreprocessNuclei( QWidget ):
    """ Preprocessing of the 3D nuclei (filters, denoising) """

    def __init__( self, viewer, mig, cfg ):
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
        super().__init__()
    
        print("********** Preprocessing nuclei signal *************")
        ## load default parameters
        rm_bg = False
        rm_bg_rad = 20
        medfilt = False
        medfilt_rad = 4
        paras = self.cfg.read_parameter_set("PreprocessNuclei")
        if paras is not None:
            if "remove_background" in paras:
                rm_bg = (paras["remove_background"].strip() == "True")
            if "median_filter" in paras:
                medfilt = (paras["median_filter"].strip() == "True")
            if "remove_background_radius" in paras:
                rm_bg_rad = int(paras["remove_background_radius"])
            if "median_filter_radius" in paras:
                medfilt_rad = int(paras["median_filter_radius"])

        layout = QVBoxLayout()
        ## median filtering option
        median_line, self.median_filtering, self.median_radius = fwid.check_value_line( "Median filtering", medfilt, "with radius ", medfilt_rad, descr="Apply median filtering to the nuclei raw image with the given radius" ) 
        layout.addLayout( median_line )
        ## remove background option
        bg_line, self.remove_background, self.remove_background_radius = fwid.check_value_line( "Remove background", rm_bg, "with radius ", rm_bg_rad, descr="Remove background with a rolling ball algorithm with the given radius" )
        layout.addLayout( bg_line )

        ## go preprocess
        apply_btn = fwid.add_button( "Apply preprocessing", btn_func=self.preprocess_go, descr="Apply the selected preprocessing steps", color=ut.get_color("go") )
        layout.addWidget( apply_btn )
        
        ## noise2void btn
        self.noise2void = fwid.add_button( "Noise2Void", self.noise2void_show, descr="Apply Noise2Void denoising" )
        self.end_n2v = fwid.add_button( "Noise2void done", self.noise2void_close, descr="Finish the noise2void step and go back to preprocessing", color=ut.get_color("done") )
        n2v_line = fwid.double_widget( self.noise2void, self.end_n2v )
        layout.addLayout( n2v_line )

        ## reset staining btn
        reset_btn = fwid.add_button( "Reset nuclei staining", btn_func=self.reset_nuclei, descr="Reset the nuclei staining to the original raw image", color=ut.get_color("reset") )
        ## done button
        done_btn = fwid.add_button( "Preprocessing done", btn_func=self.preprocess_done, descr="Finish the nuclei preprocessing", color=ut.get_color("done") )
        btn_line = fwid.double_widget( reset_btn, done_btn )
        layout.addLayout( btn_line )

        self.setLayout( layout )
    
        self.mig.prepare_segmentation_nuclei()
        ut.remove_layer( self.viewer, "nucleiStaning" )
        self.viewer.add_image( self.mig.nucstain, name="nucleiStaining", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap="blue" )
        ut.hide_color_layers( self.viewer, self.mig )

    def preprocess_go( self ):
        """ Launch the preprocessing steps """
        print("Applying preprocessing")
        self.cfg.addGroupParameter("PreprocessNuclei")
        self.cfg.addParameter("PreprocessNuclei", "remove_background_radius", int(float(self.remove_background_radius.text())))
        self.cfg.addParameter("PreprocessNuclei", "median_filter_radius", int(float(self.median_radius.text())))
        self.cfg.addParameter("PreprocessNuclei", "remove_background", self.remove_background.isChecked())
        self.cfg.addParameter("PreprocessNuclei", "median_filter", self.median_filtering.isChecked())
        self.cfg.write_parameterfile()

        if self.remove_background.isChecked():
            self.mig.preprocess_nuclei_removebg( int(float(self.remove_background_radius.text())) )
        if self.median_filtering.isChecked():
            self.mig.preprocess_nuclei_median( int(float(self.median_radius.text())) )

        ut.remove_layer( self.viewer, "nucleiStaining" )
        self.viewer.add_image( self.mig.nucstain, name="nucleiStaining", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap="blue" )
        print("Preprocessing applied")

    def noise2void_show( self ):
        """ Opens noise2void plugin """
        self.mig.prepare_nuclei()
        try:
            from napari_n2v import PredictWidgetWrapper
            self.viewer.window.add_dock_widget(PredictWidgetWrapper(self.viewer))
        except ImportError:
            ut.show_error("Please install napari-n2v to use this feature.")

    def noise2void_close( self ):
        """ Closes noise2void plugin """
        if "Denoised" in self.viewer.layers:
            ut.remove_widget(self.viewer, "Dock widget 1")
            ut.remove_layer(self.viewer, "nucleiStaining")
            self.viewer.layers["Denoised"].name = "nucleiStaining"
            self.mig.nucstain = self.viewer.layers["nucleiStaining"].data
        if "nucleiStaining" in self.viewer.layers:
            self.viewer.layers["nucleiStaining"].refresh()

    def reset_nuclei( self ):
        ut.show_info("Reset nuclei staining")
        self.mig.nucstain = None
        self.mig.prepare_segmentation_nuclei()
        ut.remove_layer( self.viewer, "nucleiStaining" )
        self.viewer.add_image( self.mig.nucstain, name="nucleiStaining", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap="blue" )

    def preprocess_done( self ):
        """ Finish the step """
        ut.remove_layer(self.viewer, "nucleiStaining")
        self.noise2void_close()
        ut.remove_widget( self.viewer, "Preprocess Nuclei" )





import fish_feats.Utils as ut
import os
from qtpy.QtWidgets import QPushButton, QCheckBox, QHBoxLayout, QVBoxLayout, QLabel, QLineEdit, QComboBox, QSpinBox, QSlider, QGroupBox, QFileDialog, QListWidget, QAbstractItemView # type: ignore from qtpy.QtCore import Qt # type: ignore
from qtpy.QtCore import Qt # type: ignore

def help_button( link, description="", display_settings=None ):
    """ Create a new Help button with given parameter """
    def show_doc():
        """ Open documentation page """
        ut.show_documentation_page( link )

    help_btn = QPushButton( "help" )
    if description == "":
        help_btn.setToolTip( "Open Fish&Feats documentation" )
        help_btn.setStatusTip( "Open Fish&Feats documentation" )
    else:
        help_btn.setToolTip( description )
        help_btn.setStatusTip( description )
    help_btn.clicked.connect( show_doc )
    if display_settings is not None:
        color = display_settings["Help button"]
        help_btn.setStyleSheet( 'QPushButton {background-color: '+color+'}' )
    else:
        color = ut.get_color( "help" )
        help_btn.setStyleSheet( 'QPushButton {background-color: '+color+'}' )
    return help_btn


def checkgroup_help( name, checked, descr, help_link, display_settings=None, groupnb=None ):
    """ Create a group that can be show/hide with checkbox and an help button """
    group = QGroupBox()
    chbox = QCheckBox( text=name )

    ## set group and checkbox to the same specific color
    if (groupnb is not None) and (display_settings is not None):
        if groupnb in display_settings:
            color = display_settings[groupnb]
            group.setStyleSheet( 'QGroupBox {background-color: '+color+'}' )
            chbox.setStyleSheet( 'QCheckBox::indicator {background-color: '+color+'}' )
    
    def show_hide():
        group.setVisible( chbox.isChecked() )

    line = QHBoxLayout()
    ## create checkbox
    chbox.setToolTip( descr )
    line.addWidget( chbox )
    chbox.stateChanged.connect( show_hide )
    chbox.setChecked( checked )
    ## create button
    if help_link is not None:
        help_btn = help_button( help_link, "", display_settings )
        line.addWidget( help_btn )
    return line, chbox, group

def group_layout( name, descr="", color=None ):
    """ Create a group layout with a name and a description """
    group = QGroupBox( name )
    if descr != "":
        group.setToolTip( descr )
        group.setStatusTip( descr )
    if color is not None:
        group.setStyleSheet( 'QGroupBox {background-color: '+color+'}' )
    layout = QVBoxLayout()
    return group, layout

def add_multiple_list( label, descr="" ):
    """ List interface, with possibility of multiple selection """
    list_widget = QListWidget()
    list_widget.setSelectionMode(QAbstractItemView.MultiSelection)
    layout = QVBoxLayout()
    layout.addWidget(QLabel(label))
    layout.addWidget(list_widget)
    if descr != "":
        list_widget.setToolTip( descr )
    return layout, list_widget

def add_button( btn, btn_func, descr="", color=None ):
    """ Add a button connected to an action when pushed """
    cbtn = QPushButton( btn )
    if btn_func is not None:
        cbtn.clicked.connect( btn_func )
    if descr != "":
        cbtn.setToolTip( descr )
    else:
        cbtn.setToolTip( "Click to perform action" )
    if color is not None:
        cbtn.setStyleSheet( 'QPushButton {background-color: '+color+'}' )
    return cbtn

def double_button( btna, btnb ):
    """ Create a line with two buttons """
    line = QHBoxLayout()
    line.addWidget( btna )
    line.addWidget( btnb )
    return line

def add_label( lab, descr="" ):
    """ Create a label object """
    label = QLabel( lab )
    if descr != "":
        label.setToolTip( descr )
        label.setStatusTip( descr )
    return label


def label_button( btn, btn_func, label="", descr="", color=None ):
    """ Create a line with a label and a button """
    line = QHBoxLayout()
    lab = QLabel()
    lab.setText( label )
    if descr != "":
        lab.setToolTip( descr )
        lab.setStatusTip( descr )
    btn = add_button( btn, btn_func, descr, color )
    line.addWidget( btn )
    line.addWidget( lab )
    return line, btn, lab


def line_button_help( btn, btn_func, descr="", help_link=None, color=None ):
    """ Create a line with a button and an help button """
    line = QHBoxLayout()
    btn = add_button( btn, btn_func, descr, color )
    line.addWidget( btn )
    if help_link is not None:
        help_btn = help_button( help_link, "Open online documentation", None )
        line.addWidget( help_btn )
    return line, btn

def list_line( label, descr="", func=None ):
    """ Create a layout line with a choice list to edit (non editable name + list part ) """
    line = QHBoxLayout()
    ## Value name
    lab = QLabel()
    lab.setText( label )
    line.addWidget( lab )
    if descr != "":
        lab.setToolTip( descr )
        lab.setStatusTip( descr )
    ## Value editable part
    value = QComboBox()
    line.addWidget( value )
    if func is not None:
        value.currentIndexChanged.connect( func )
    return line, value

def add_value( default_value, descr="" ):
    """ Editable value """
    value = QLineEdit()
    value.setText( str(default_value) )
    if descr != "":
        value.setToolTip( descr )
    return value

def double_value_line( labela, default_valuea, labelb, default_valueb, descr="" ):
    """ Layout with two values to edit """
    line = QHBoxLayout()
    ## first
    laba = QLabel()
    laba.setText( labela )
    line.addWidget( laba )
    valuea = QLineEdit()
    valuea.setText( str(default_valuea) )
    line.addWidget( valuea )
    ## second
    labb = QLabel()
    labb.setText( labelb )
    line.addWidget( labb )
    valueb = QLineEdit()
    valueb.setText( str(default_valueb) )
    line.addWidget( valueb )
    if descr != "":
        laba.setToolTip( descr )
        labb.setToolTip( descr )
    return line, valuea, valueb

def value_line( label, default_value, descr="" ):
    """ Create a layout line with a value to edit (non editable name + value part ) """
    line = QHBoxLayout()
    ## Value name
    lab = QLabel()
    lab.setText( label )
    line.addWidget( lab )
    if descr != "":
        lab.setToolTip( descr )
    ## Value editable part
    value = QLineEdit()
    value.setText( str(default_value) )
    line.addWidget( value )
    return line, value

def spinner_line( name, minval, maxval, step, value, changefunc=None, descr="" ):
    """ Line with a spinbox """
    line = QHBoxLayout()
    ## add name if any
    if name is not None:
        lab = QLabel()
        lab.setText( name )
        line.addWidget( lab )
    ## add spinbox
    spinner = QSpinBox()
    spinner.setRange( minval, maxval )
    spinner.setSingleStep(1)
    spinner.setValue( int(value) )
    if changefunc is not None:
        spinner.valueChanged.connect( changefunc )
    if descr != "":
        spinner.setToolTip( descr )
    line.addWidget( spinner )
    return line, spinner

def slider_line( name, minval, maxval, step, value, show_value=False, slidefunc=None, descr="", div=1 ):
    """ Line with a text and a slider """
    line = QHBoxLayout()
    ## add name if any
    if name is not None:
        lab = QLabel()
        lab.setText( name )
        line.addWidget( lab )
    ## add slider
    slider =  QSlider( Qt.Horizontal )
    slider.setMinimum( int(minval*div) )
    slider.setMaximum( int(maxval*div) )
    slider.setSingleStep( int(step*div) )
    slider.setValue( int(value*div) )
    if slidefunc is not None:
        slider.valueChanged.connect( slidefunc )
    if descr != "":
        slider.setToolTip( descr )
        lab.setToolTip( descr )
    if show_value:
        lab = QLabel(""+str(value*1.0))
        line.addWidget( lab )
        slider.valueChanged.connect( lambda: lab.setText( ""+str(slider.value()*1.0/div) ) )
    line.addWidget( slider )
    return line, slider

def double_widget( wida, widb ):
    """ Create a line layout with the two widgets """
    line = QHBoxLayout()
    line.addWidget( wida )
    line.addWidget( widb )
    return line

def check_value_line( name, checked,  value_name, value_default, descr="" ):
    """ Line with value visible only if checked """
    line = QHBoxLayout()
    cbox = QCheckBox( text=name )
    cbox.setChecked( checked )
    line.addWidget( cbox )
    lab = QLabel()
    lab.setText( value_name )
    line.addWidget( lab )
    value = QLineEdit()
    value.setText( str(value_default) )
    line.addWidget( value )
    if descr != "":
        lab.setToolTip( descr )
    cbox.stateChanged.connect( lambda state: value.setVisible( state == Qt.Checked ) )
    cbox.stateChanged.connect( lambda state: lab.setVisible( state == Qt.Checked ) )
    value.setVisible( checked )
    lab.setVisible( checked )
    return line, cbox, value

def add_check( check, checked, check_func=None, descr="" ):
    """ Add a checkbox with set parameters """
    cbox = QCheckBox( text=check )
    cbox.setToolTip( descr )
    if check_func is not None:
        cbox.stateChanged.connect( check_func )
    cbox.setChecked( checked )
    return cbox

def file_line( title, default_path="", dial_msg="Choose file", filetype="All (*)", descr="" ):
    """ interface line with browse btn and current path """
    line = QHBoxLayout()
    lab = add_label( title, descr )
    line.addWidget( lab )
    value = QLineEdit()
    value.setText( default_path )
    #value.setMaximumWidth(20)
    line.addWidget( value )
    btn = QPushButton( "Browse" )
    btn.clicked.connect( lambda: browse_file( value, dial_msg, filetype, default_path ) )
    line.addWidget( btn )
    return line, value

def dir_line( title, default_path="", dial_msg="Choose directory", descr="" ):
    """ interface line with browse btn and current path """
    line = QHBoxLayout()
    lab = add_label( title, descr )
    line.addWidget( lab )
    value = QLineEdit()
    value.setText( default_path )
    line.addWidget( value )
    btn = QPushButton( "Browse" )
    btn.clicked.connect( lambda: browse_dir( value, dial_msg, default_path ) )
    line.addWidget( btn )
    return line, value

def browse_file( value, dial_msg, filetype, default_path ):
    """ Open a file dialog to select a file """
    filepath = file_dialog( dial_msg, filetype, directory=default_path )
    if filepath is not None:
        value.setText( filepath )   

def browse_dir( value, dial_msg, default_path ):
    """ Open a file dialog to select a directory """
    dirpath = dir_dialog( dial_msg, directory=default_path )
    if dirpath is not None:
        value.setText( dirpath )

def dir_dialog( title, directory=None ):
    """ Opens a dialog to select a folder """
    if directory is None:
        directory = ''
    else:
        directory = str( os.path.dirname(directory) )
    file_dialog = QFileDialog()
    file_dialog.setFileMode(QFileDialog.Directory)
    #file_dialog.setOption(QFileDialog.ShowDirsOnly, True)
    file_dialog.setWindowTitle(title)
    file_dialog.setDirectory(directory)
    if file_dialog.exec_():
        filepath = file_dialog.selectedFiles()[0]
    else:
        filepath = None
    return filepath 

def file_dialog( title, filetypes, directory=None ):
    """ Open a file dialog to select a file """
    if directory is None:
        directory = ''
    else:
        directory = str( os.path.dirname(directory) )
    file_dialog = QFileDialog()
    file_dialog.setFileMode(QFileDialog.ExistingFile)
    file_dialog.setNameFilter(filetypes)
    file_dialog.setWindowTitle(title)
    file_dialog.setDirectory(directory)
    if file_dialog.exec_():
        filepath = file_dialog.selectedFiles()[0]
        #if not filepath.endswith('.tif'):
        #    raise ValueError("Selected file is not a .tif file")
    else:
        filepath = None
    return filepath 



import numpy as np
import os, csv, time
import glob
import matplotlib as plt
from napari.utils.translations import trans
from napari.utils.history import get_save_history, update_save_history 
from qtpy.QtWidgets import QFileDialog
from magicgui.widgets import TextEdit
from magicgui.widgets import Table
import vispy.color
from napari.utils import notifications as nt
from napari.utils import progress # type: ignore
import webbrowser
from packaging.version import Version
import napari
from skimage.segmentation import find_boundaries
from skimage.morphology import binary_closing, binary_dilation, disk
from skimage.measure import regionprops_table
import gc

"""
    Set of common functions for Fish&Feats pipeline
"""

#### Napari plugin functions
def show_info(message):
    """ Display information window in napari """
    nt.show_info(message)

def show_warning(message):
    """ Display a warning in napari (napari function show_warning doesn't work) """
    mynot = nt.Notification(message, nt.NotificationSeverity.WARNING)
    nt.notification_manager.dispatch(mynot)

def show_error(message):
    """ Display an error in napari (napari function show_error doesn't work) """
    mynot = nt.Notification(message, nt.NotificationSeverity.ERROR)
    nt.notification_manager.dispatch(mynot)

def show_debug(message):
    """ Display an info for debug in napari (napari function show_debug doesn't work) """
    mynot = nt.Notification(message, nt.NotificationSeverity.DEBUG)
    nt.notification_manager.dispatch(mynot)

def show_progress( viewer, show ):
    """ Show.hide the napari activity bar to see processing progress """
    viewer.window._status_bar._toggle_activity_dock( show )

def start_progress( viewer, total, descr=None ):
    """ Start the progress bar """
    show_progress( viewer, True)
    progress_bar = progress( total )
    if descr is not None:
        progress_bar.set_description( descr )
    return progress_bar

def close_progress( viewer, progress_bar ):
    """ Close the progress bar """
    progress_bar.close()
    show_progress( viewer, False)

def get_time():
    """ Returns current time"""
    return time.time()

def show_duration(start_time, header=None):
    if header is None:
        header = "Processed in "
    #show_info(header+"{:.3f}".format((time.time()-start_time)/60)+" min")
    print(header+"{:.3f}".format((time.time()-start_time)/60)+" min")


def show_documentation_page(page):
    """ Open the wiki documentation """
    webbrowser.open_new_tab("https://gletort.github.io/FishFeats/"+page)

def get_color( obj ):
    """ Returns the color of the object """
    if obj == "help":
        return "#665071"
    if obj == "save":
        return "#455873" #"#5F6BFA"
    if obj == "load":
        return "#1D3658" #"#5F6BFA"
    if obj == "done":
        return "#7A0FF5"
    if obj == "go":
        return "#2C78B692"
    if obj == "reset":
        return "#0E121492"
    if obj == "light":
        return "#B1B7F8"
    if obj == "group1":
        return "#384f54" #"#4c617f"
    if obj == "group2":
        return "#363650"
    if obj == "group3":
        return "#212a37"
    return "#000000"

def dialog_filename():
    dialog = QFileDialog()
    hist = get_save_history()
    dialog.setHistory(hist)
    dialog.setFileMode(QFileDialog.ExistingFile)
    dialog.setDirectory(hist[0])
    if dialog.exec_():
        filename = dialog.selectedFiles()
    if filename:
        return filename[0]
    else:
        return None

def do_nothing(viewer):
    return

def update_history(mdir):
    update_save_history(mdir)

def quantiles(img):
    """ Returns quantiles for contrast setting """
    return tuple(np.quantile(img, [0.01, 0.9999]))

def main_shortcuts(viewer):
    """ Open a new text window with the main shortcuts and help """
    blabla = TextEdit()
    blabla.name = "Fish&Feats shortcuts"
    text = "-------------------------- Fish&Feats -------------------------- \n"
    text += view_shortcuts()
    text += labels_shortcuts( level = 1 )
    text += association_shortcuts()
    text += rnas_shortcuts()
    text += pos3d_shortcuts()
    text += classify_shortcuts()
    blabla.value = text
    blabla.show()
    return blabla

def help_shortcut( step ):
    """ Main shortcuts to display for each step"""
    if step == "view":
        help_text = "<h> show/hide this message \n"
        help_text += "<Ctrl-h> open main shortcuts list \n"
        help_text += "<F1>, <F2>... show/hide the first/second... layer \n"
        return help_text
    if step == "pos3d":
        help_text = "<Right-click> on a cell to get its label in the cell_label parameter \n"
        help_text = help_text + "<Control+left-click> when the viewer is at the correct Z position to set it in the place_at_z parameter \n"
        return help_text
    if step == "classify":
        text = "Press <i>/<d> to increase/decrease the filling value (category)\n"
        text += "<Control+Left-click> on a cell to assign it the current filling value \n"
        text += "<Right-click> on a cell to select it and get its value \n" 
        text += "<F1>, <F2>... show/hide the first/second... layer \n"
        return text
    if step == "projection":
        text = "<h> to show/hide this message \n"
        text += "<F1>, <F2>... show/hide the first/second... layer \n"
        text += "Junctions are segmented in 2D and then back-projected to their Z location \n"
        text += "Load a file of the 2D projection or calculate it now \n"
        return text

def scale_shortcuts():
    """ Help message for metadata interface """
    help_text = "Check and correct the scaling information \n"
    help_text += "Choose the color channels in which the junction and nuclei staining are (the channel numbers should be the same as the layer names originalChannel*) \n"
    help_text += "\'load previous\' option will load all saved files (in results folder) for the current images \n"
    return help_text

def view_shortcuts():
    """ Viewing shortcuts list """
    text = "---------- Main viewing shortcuts, always active \n"
    text += "  <Ctrl+Y> switch 2D/3D view mode \n"
    text += "  <Ctrl-v> in 3D, switch vispy mode ON/OFF \n"
    text += "  <v> show/hide current selected layer \n"
    text += "  <Ctrl+R> reset view \n"
    text += "  <left arrow> go to previous frame \n"
    text += "  <right arrow> go to next frame \n"
    text += "  <Ctrl+G> switch Grid/Overlay view mode \n"
    text += "  <g> show/hide fish grid \n"
    text += help_shortcut("view")
    text += "\n"
    return text

def classify_shortcuts():
    """ Shortcuts for cell classification """
    text = " ---------- Cell classification shortcuts (the feature layer should be active) \n"
    text += help_shortcut("classify")
    text += "\n"
    return text

def pos3d_shortcuts():
    """ Shortcuts for 3D positionning of cells """
    text = " ---------- 3D positionning shortcuts, active on CellContours layer \n"
    text = text + help_shortcut("pos3d")
    text += "\n"
    return text

def labels_shortcuts( level=1 ):
    text = ""
    if level >= 1:
        text += " ---------- labels options (for cells and nuclei) \n"
        text += "  <2> painting brush mode \n"
        text += "  <3> fill mode \n"
        text += "  <4> select label mode \n"
        text += "  <5> Swith to zoom/moving mode \n"
        text += "  <[> or <]> increase/decrease the paint brush size \n"
    text += "  <p> activate/deactivate preserve labels option \n"
    text += "  <m> to set current label to max+1 value \n"
    text += "  <l> to show/hide cell/nuclei labels. "
    if level >= 1:
        text += "It creates a new layer with the object label (number) around each object position. \n"
    else:
        text += "\n"
    text += "  <Ctrl-c>/<Ctrl-d> increase/decrease label contour size (0=full)\n"
    text += "  <Control+left click> from one label to another to merge them (the label kept will be the last one) \n"
    text += "  <Control+right-click> on a label to erase it \n" 
    if level >= 1:
        text += " Double <Left-click> to zoom in \n"
        text += "\n"
        text += "For 3D: \n"
        text += "In 3D, most label actions wont work if Vispy perspective is ON. Switch it off with 'Ctrl-v' before.\n"
        text += "If n_edit_dim is set on 3 (top left panel), edition will affect all or several z (slices) \n"
        text += "If n_edit_dim is set on 2, edition will only affect the active slice \n"
        text += "\n"
    return text

def association_shortcuts():
    """ Shorcuts for junction-nuclei association manual correction """
    help_text = " ---------- Association options (link nucleus to corresponding cell) \n"
    help_text += "<Double-click> to select a nucleus to correct \n"
    help_text = help_text + "<Right-click> to choose the cell to associate with \n"
    help_text = help_text + "<c> to apply current association \n"
    help_text = help_text + "<l> to show/hide cell labels \n"
    help_text = help_text + "<s> to synchronize junctions and nuclei view \n"
    help_text = help_text + "<u> to unsynchronize junctions and nuclei view \n"
    help_text += "  <Ctrl-c>/<Ctrl-d> increase/decrease NUCLEI label contour (0: full) \n"
    help_text += "  <Alt-c>/<Alt-d> increase/decrease JUNCTIONS label contour (0: full) \n"
    help_text += "\n"
    return help_text


def rnas_shortcuts():
    text = " -------------- Points options (for RNA dots) \n"
    text += "  <2> add points mode \n"
    text += "  <3> select points mode \n"
    text += "  <4> Pan/zoom \n"
    text += "  <A> select all visible points \n" 
    text += "  <v> to show/hide cell contours \n"
    text += "  <Right-click> on a cell to set the assignement value \n"
    text += "  <c> assign selected points to the current assignement value \n"
    text += "  <u> select all unsassigned points \n"
    text += "  <s> select points with current assignement value\n"
    text += "  <o> show only selected points \n"
    text += "  <r> reset, show all points \n"
    text += "  <l> switch show only selected cell/all cells \n"
    text += "  <Ctrl-v> switch Vispy perspective mode on/off \n"
    text += "\n"
    return text

def helpHeader(viewer, layname = None):
    text = "------------- ShortCuts ---------------- \n"
    if layname is not None:
        text = text + "If layer "+layname+" is active: \n"
    text = text + "<h> show/hide this help message \n"
    return text

def removeOverlayText(viewer):
    viewer.text_overlay.text = trans._("")
    viewer.text_overlay.visible = False

def showOverlayText(viewer, text, size=13, col="white" ):
    viewer.text_overlay.text = trans._(text)
    viewer.text_overlay.visible = True
    viewer.text_overlay.font_size = size
    viewer.text_overlay.color = col
    viewer.text_overlay.position = "top_left"

def showHideOverlayText(viewer, vis=None):
    """ Show on/off overlay text """
    if vis is None:
        viewer.text_overlay.visible = not viewer.text_overlay.visible
    else:
        viewer.text_overlay.visible = vis

def get_layer(viewer, layname):
    if layname in viewer.layers:
        return viewer.layers[layname]
    return None

def set_active_layer(viewer, layname):
    """ Set the current Napari active layer """
    if layname in viewer.layers:
        viewer.layers.selection.active = viewer.layers[layname]

def view_3D( viewer ):
    """ Put the viewer in 3D mode """
    viewer.dims.ndisplay = 3

def remove_all_layers( viewer ):
    """ Remove all open layers """
    viewer.layers.clear()

def remove_layer(viewer, layname):
    if layname in viewer.layers:
        viewer.layers.remove(layname)

def remove_all_widget( viewer ):
    """ Remove all widgets """
    viewer.window.remove_dock_widget("all")
    
def list_widgets( viewer ):
    """ List all open widgets """
    if not version_napari_above("0.6.1"):
        return list(viewer.window._dock_widgets.keys())
    return list(viewer.window.dock_widgets.keys())

def has_widget( viewer, widname ):
    """ Check if given widget is open """
    if not version_napari_above("0.6.1"):
        return widname in viewer.window._dock_widgets
    return widname in viewer.window.dock_widgets

def remove_widget(viewer, widname):
    """ Remove a widget from the viewer """
    if not version_napari_above("0.6.1"):
        ## functions changed in recent version
        if widname in viewer.window._dock_widgets:
            wid = viewer.window._dock_widgets[widname]
            wid.setDisabled(True)
            try:
                wid.disconnect()
            except:
                #print("Widget "+widname+" deleted but not disconnected (pyside2)")
                pass
            del viewer.window._dock_widgets[widname]
            wid.destroyOnClose()
    else:
        if widname in viewer.window.dock_widgets:
            if not version_napari_above("0.6.2"):
                wid = viewer.window.dock_widgets[widname]
                wid.setDisabled(True)
                try:
                    wid.disconnect()
                except:
                    #print("Widget "+widname+" deleted but not disconnected (pyside2)")
                    pass
                del viewer.window._dock_widgets[widname]
                wid.destroyOnClose()
            else:
                viewer.window.remove_dock_widget( viewer.window.dock_widgets[widname] )

def hide_color_layers(viewer, mig):
    """ Hide all the originalChannel layers """
    for chan in range(mig.nbchannels):
        viewer.layers["originalChannel"+str(chan)].visible = False

def show_layer(viewer, chan):
    """ Show the color layer nchan """
    viewer.layers["originalChannel"+str(chan)].visible = True


def scale_layer( viewer, layer, scale_tuple ):
    """ Scale the layer to the scale_tuple """
    if layer in viewer.layers:
        viewer.layers[layer].scale = scale_tuple
        viewer.layers[layer].refresh()

### color maps
def colormaps():
    mixor = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.9, 0.5, 0.6]])
    mixgr = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.4, 0.75, 0.55]])
    mix5 = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.45, 0.35, 0.25]])
    mixbl = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.55, 0.45, 0.75]])
    mix4 = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.65, 0.7, 0.85]])
    return ["red", "green", "blue", "yellow", "cyan", "gray", "magenta", "bop orange", "bop purple", "bop blue", mixor, mix5, mixbl, mixgr, mix4]


def colormapname( i ):
    colorsmaps = colormaps()
    ncolmaps = len(colorsmaps)
    return colorsmaps[(i%ncolmaps)]
    
def create_labelmap():
    cmap = plt.cm.gist_rainbow
    tabmap = np.arange(cmap.N)
    np.random.shuffle(tabmap)
    my_cmap = cmap(tabmap)
    alph = np.repeat(1, cmap.N)
    alph[0] = 0
    my_cmap[:, -1] = alph
    return my_cmap

#### opening/writing
def openims(imagepath, verbose=True):
    """ Open ims image """
    from imaris_ims_file_reader.ims import ims
    img = ims(imagepath, squeeze_output=True )
    image = np.array(img[0])
    scaleXY = img.resolution[1]
    if img.resolution[2] != scaleXY:
        print("Warning, scale is not the same in X and Y, not implemented yet")
    scaleZ = img.resolution[0]
    if verbose:
        print("Initial image shape: "+str(image.shape))    ## Y et X sont inverses, Z, Y, X
    nchan = img.Channels
    names = []
    try:
        if nchan > 0:
            for i in range(nchan):
                name = img.read_attribute("DataSetInfo/Channel "+str(i), 'DyeName')
                names.append(name)
    except:
        names = []
    img = None

    return (np.squeeze(image), scaleXY, scaleZ, names)

def writeims(imagepath, img, verbose=True):
    ## libraries pb, must install hdf5, ImarisWriter
    from PyImarisWriter import PyImarisWriter as pw
    imshape = img.shape
    image_size = pw.ImageSize(x=imshape[2], y=imshape[1], z=imshape[0], c=1, t=1)
    dimension_sequence = pw.DimensionSequence('z', 'y', 'x', 'c', 't')
    block_size = image_size
    sample_size = pw.ImageSize(x=1, y=1, z=1, c=1, t=1)
    output_filename = 'outifle.ims'

    options = pw.Options()
    options.mNumberOfThreads = 12
    options.mCompressionAlgorithmType = pw.eCompressionAlgorithmGzipLevel2
    options.mEnableLogProgress = True

    application_name = 'PyImarisWriter'
    application_version = '1.0.0'

    #callback_class = MyCallbackClass()
    converter = pw.ImageConverter("uint8", image_size, sample_size, dimension_sequence, block_size,
                                  output_filename, options, application_name, application_version, None)

    num_blocks = image_size / block_size

    block_index = pw.ImageSize()
    for c in range(num_blocks.c):
        block_index.c = c
        for t in range(num_blocks.t):
            block_index.t = t
            for z in range(num_blocks.z):
                block_index.z = z
                for y in range(num_blocks.y):
                    block_index.y = y
                    for x in range(num_blocks.x):
                        block_index.x = x
                        if converter.NeedCopyBlock(block_index):
                            converter.CopyBlock(img, block_index)

    adjust_color_range = True
    image_extents = pw.ImageExtents(0, 0, 0, image_size.x, image_size.y, image_size.z)
    parameters = pw.Parameters()
    parameters.set_value('Image', 'ImageSizeInMB', 2400)
    parameters.set_value('Image', 'Info', 'Results Title')
    parameters.set_channel_name(0, 'My Channel 1')
    time_infos = [datetime.today()]
    color_infos = [pw.ColorInfo() for _ in range(image_size.c)]
    color_infos[0].set_color_table(configuration.mColor_table)

    converter.Finish(image_extents, parameters, time_infos, color_infos, adjust_color_range)

    converter.Destroy()
    print('Wrote file')

def writePng( img, imgname ):
    """ Save an image as png """
    import matplotlib.image
    matplotlib.image.imsave( imgname, img )

def writeTif(img, imgname, scaleXY, scaleZ, imtype):
    import tifffile
    ### channels and z
    if (scaleZ >= 0) and (len(img.shape)>3):
        arr = np.array( img, dtype=imtype )
        ## exhange C and Z axis for Fiji order
        arr = np.moveaxis( arr, 0, 1 )
        tifffile.imwrite(imgname, arr, imagej=True, resolution=[1./scaleXY, 1./scaleXY], dtype=imtype, metadata={'spacing': scaleZ, 'unit': 'um', 'axes': 'ZCYX'})
        return
    ### 2D
    if scaleZ < 0:
        tifffile.imwrite(imgname, np.array(img, dtype=imtype), imagej=True, dtype=imtype, resolution=[1./scaleXY, 1./scaleXY], metadata={'unit': 'um', 'axes': 'YX'})
    #### 3D
    else:
        tifffile.imwrite(imgname, np.array(img, dtype=imtype), imagej=True, dtype=imtype, resolution=[1./scaleXY, 1./scaleXY], metadata={'spacing': scaleZ, 'unit': 'um', 'axes': 'ZYX'})

def get_scale_of(racine, c):
    """ read scaling value from xml metadata """
    try:
        part = racine.xpath("//Distance [@Id = '%s']" % c)
        for neighbor in part[0].iter('Value'):
            pixel_in_meters = float(neighbor.text)
    except:
        pixel_in_meters = -1
    return pixel_in_meters*1000000

def get_fluo_names(racine):
    names = []
    try:
        for i in range(10):
            chans = racine.xpath("//Channel [@Id = \"Channel:%s\"]" % i)
            if len(chans) > 0:
                for j in range(len(chans)):
                    dyename = chans[j].findall("DyeName")
                    if len(dyename) > 0:
                        names.append(dyename[0].text)
    except:
        names = []
    return names

def openczi(imagepath, scene=0, verbose=True):
    """ Open czi image """
    import czifile
    from lxml import etree
    czi = czifile.CziFile(imagepath)
    image = czi.asarray()
    if verbose:
        print("Initial image shape: "+str(image.shape))

    if image.ndim >= 8:
        image = image[:,scene,:,:,:,:,:,:]  ## si contient 2 scenes

    # get scale
    root = etree.fromstring(czi.metadata())
    scaleXY = get_scale_of(root, "X")
    scaleXYcheck = get_scale_of(root, "Y")
    if scaleXY != scaleXYcheck:
        print("Warning, scale not the same in X and in Y, not implemented yet")
    scaleZ = get_scale_of(root, "Z")

    fluonames = get_fluo_names(root)
    return (np.squeeze(image), scaleXY, scaleZ, fluonames)

def getBestZ( img, verbose=True ):
    proj = np.sum(img, axis=1)  ## sum les x
    proj = np.sum(proj, axis=1) ## sum les y
    zref = np.where(proj==np.max(proj))[0][0]
    if verbose:
        print("Reference z: "+str(zref))
    return zref

def get_filename():
    try:
        from tkinter import Tk
        from tkFileDialog import askopenfilenames
    except:
        from tkinter import Tk
        from tkinter import filedialog

    Tk().withdraw() # we don't want a full GUI, so keep the root window from appearing
    filenames = filedialog.askopenfilenames() # show an "Open" dialog box and return the path to the selected file
    return filenames

def get_filelist(pattern):
    return glob.glob(pattern)


def opentif(imagepath, verbose=True):
    import tifffile as tif
    img = tif.TiffFile(imagepath)
    metadata = img.imagej_metadata
    names = []
    scaleXY = 1
    scaleZ = 1
    try:
        if metadata is not None:
            info = img.imagej_metadata["Info"]
            for i in range(10):
                ind = info.find("channels_channel_"+str(i)+"_dyename")
                if ind >= 0:
                    keep = info[ind:]
                    indend = keep.find("\n")
                    names.append(info[ind+28:ind+indend].strip())
     
            metadata = (img.imagej_metadata['Info']).splitlines()
            scaleXY = float(metadata[-4].split()[2])*1000000
            scaleZ = float(metadata[-2].split()[2])*1000000
    except:
        scaleXY = 0.25
        scaleZ = 1
    image = img.asarray()
    img.close()
    return image, scaleXY, scaleZ, names

def opentif_nonames(imagepath, verbose=True):
    import tifffile as tif
    img = tif.TiffFile(imagepath)
    metadata = img.imagej_metadata
    #print(metadata)
    scaleXY = 1
    scaleZ = 1
    names = []
    if metadata is not None:
        try:
            info = metadata["Info"]
            if info is not None: 
                metadatas = (info).splitlines()
                scaleXY = float(metadatas[-4].split()[2])*1000000
                scaleZ = float(metadatas[-2].split()[2])*1000000
        except:
            metadatas = None
        try:
            if metadata['spacing'] is not None:
                scaleZ = float(metadata['spacing'])
        except:
            metadatas = None
        try:
            if metadata['physicalsizex'] is not None:
                scaleXY = float(metadata['physicalsizex'])
        except:
            metadatas = None
            #print(info)
    image = img.asarray()
    img.close()
    return image, scaleXY, scaleZ, names

def arrange_dims(image, verbose=True):
    ## test if there is a channel dimension. If yes, put it first in the order
    if len(image.shape)>3:
        chanpos = np.where(image.shape==np.min(image.shape))[0][0]
        if chanpos != 0:
            image = image.swapaxes(chanpos, 0)

    ## if there is no channel dimension, add one to always have one
    if len(image.shape)<=3:
        image = np.expand_dims(image, axis=0)

    if verbose:
        print("Image dimensions: "+str(image.shape))

        return image

def extract_names(imagepath, subname="results"):
    """ From the image name, get specific paths """
    imgname = os.path.splitext(os.path.basename(imagepath))[0]
    imgdir = os.path.dirname(imagepath)
    resdir = os.path.join(imgdir, subname)
    if not os.path.exists(resdir):
        os.makedirs(resdir)
    return imgname, imgdir, resdir

def open_image(imagepath, verbose=True):
    """ Open an image from different format """
    ext = os.path.splitext(imagepath)[1]
    if ext == ".ims":
        return openims(imagepath, verbose)
    elif ext == ".czi":
        return openczi(imagepath, verbose)
    elif (ext == ".tif") or (ext == ".tiff"):
        return opentif_nonames(imagepath, verbose)
    else: 
        print("Image format not implemented yet")

def get_table_header(filepath):
    """ read the header of the csv file """
    with open(filepath, 'r') as infile:
        csvreader = csv.DictReader(infile)
        return csvreader.fieldnames

def show_table(results):
    """ Show the results in a table """
    tab = {}
    for k in results[0].keys():
        tab[k] = []
        for row in results:
            tab[k].append(row[k])
    Table(tab).show()

def load_table(filepath, column_names, verbose=True):
    """ Load a table results from a .csv file - Get only columns that are within column_names """
    res  = []
    with open(filepath, 'r') as infile:
        csvreader = csv.DictReader(infile)
        for row in csvreader:
            cres = []
            for col in column_names:
                cres.append(float(row[col]))
            res.append(cres)
    #print(res)
    return res

def load_dictlist(filepath, verbose=True):
    """ Load a list of dict results from a .csv file """
    res  = []
    with open(filepath, 'r') as infile:
        csvreader = csv.DictReader(infile)
        for row in csvreader:
            res.append(row)
    #print(res)
    return res

def load_dict_int(filepath, columns, verbose=True):
    """ Load a dict results from a .csv file, put value as int """
    res  = {}
    for col in columns:
        res[col] = []
    with open(filepath, 'r') as infile:
        csvreader = csv.DictReader(infile)
        for row in csvreader:
            for col in columns:
                res[col].append(int(row[col]))
    return res

def strip_keys( indict ):
    """ Ensures that all keys doens't have unnecessary spaces"""
    outdict = {}
    for key, value in indict.items():
        skey = key.strip()
        outdict[skey] = value
    return outdict

def write_dict(resfilepath, resdict):
    """ Write the content of resdict into a csv file """
    if isinstance(resdict, dict):
        with open(resfilepath, 'w') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(resdict.keys())
            writer.writerows(zip(*resdict.values()))
    if isinstance(resdict, list):
        ## get list of possible keys
        allkeys = []
        for row in resdict:
            if isinstance(row, dict):
                keys = row.keys()
                for key in keys:
                    if key not in allkeys:
                        allkeys.append(key)
        with open(resfilepath, 'w') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(allkeys)
            for row in resdict:
                wrow = []
                for key in allkeys:
                    if key not in row.keys():
                        wrow.append( -99999 )
                    else:
                        wrow.append(row[key])
                writer.writerow( wrow )

def select_slices(image, bestz, scaleZ, zmargin=10, rm_empty=True, verbose=True):
    """ Keep only slices around bestz and slices that contain signal """
    
    dz = int(zmargin/scaleZ)    ## keep 10 microns under and above reference z
    limLz = max(bestz-int(dz),0)  
    limHz = min(bestz+dz, (image[0,:,0,0].size))
    if verbose:
        print("Z limits "+str(limLz)+" "+str(limHz))

    ## Remove slices with nearly nothing
    if rm_empty:
        simage = np.sum(image, axis=0)
        zsimage = np.sum(simage, axis=1)
        zsimage = np.sum(zsimage, axis=1)
        empty = zsimage<np.mean(zsimage)-1*np.std(zsimage)
        while empty[limLz] and limLz<limHz:
            limLz += 1
        while empty[limHz-1] and limHz>limLz:
            limHz = limHz - 1

    if verbose:
        print("Final Z limits "+str(limLz)+" "+str(limHz))

    image = image[:,limLz:limHz,:,:]
    bestz = bestz-limLz
    if verbose:
        print('Global reference z: '+str(bestz))
    return image, bestz

def normalize_img(img, newmax=1):
    """ Return a normalised version of img """
    quantiles = np.quantile(img, [0.01,0.999])
    resimg = np.clip(img, quantiles[0], quantiles[1])
    resimg = (resimg*1.0-quantiles[0])/(quantiles[1]-quantiles[0])*newmax
    return resimg

#### Handle versions of napari
def version_napari_above( compare_version ):
    """ Get the current version of napari """
    return Version(napari.__version__) > Version(compare_version)

def add_point_layer( viewer, pts, colors, layer_name, mig, size=7, pts_properties=None ):
    """ Add a points layer to the viewer """
    if not version_napari_above("0.4.19"):
        points_layer = viewer.add_points( pts,
            face_color=colors, 
            size = size, 
            edge_width=0, 
            properties = pts_properties,
            name=layer_name,
            scale=(mig.scaleZ, mig.scaleXY, mig.scaleXY),
            out_of_slice_display = True, 
            blending="additive"
            )
    else:
        points_layer = viewer.add_points( pts,
            face_color=colors, 
            size = size, 
            border_width=0, 
            properties = pts_properties,
            name=layer_name,
            scale=(mig.scaleZ, mig.scaleXY, mig.scaleXY),
            out_of_slice_display = True, 
            blending="additive"
            )

### Labels edition

def neighbor_labels(img, lab, olab):
    """ Check if the two labels are neighbors or not """
    
    flab = find_boundaries(img==lab)
    folab = find_boundaries(img==olab)
    return np.sum(np.logical_and(flab, folab))>0

def where_to_modif(img, startlab, endlab):
    """ Merge two labels in one slice (2D)"""
    tmp = np.copy(img)
    initlab = tmp==startlab
    tmp[initlab] = endlab
    joinlab = (tmp==endlab)
    footprint = disk(radius=4)
    joinlab = binary_closing(joinlab, footprint)
    tomodif = ((tmp==0)*joinlab+initlab)
    return tomodif

def merge_labels(layer, frame, startlab, endlab, extend=10):
    """ Merge the two given labels """
    img = layer.data
    if frame is not None:
        img = layer.data[frame,]
    if not neighbor_labels(img, startlab, endlab):
        print("Labels not touching, I refuse to merge them")
        return
    
    if layer.ndim == 2:
        tomodif = where_to_modif(layer.data, startlab, endlab)
        layer.data[tomodif] = endlab
    else:
        for ind, islice in enumerate(layer.data):
            tomodif = where_to_modif(islice, startlab, endlab)
            layer.data[ind,][tomodif] = endlab
    layer.refresh()

def insideBoundingBox(pt, bbox, margin=0):
    """ Test if a point is inside a bounding box with a given extra margin"""
    if pt[0] < bbox[0]-margin:
        return False
    if pt[0] > bbox[2]+margin:
        return False
    if pt[1] < bbox[1]-margin:
        return False
    if pt[1] > bbox[3]+margin:
        return False
    return True

def cropBbox( img, bbox ):
    """ Crop the image following bounding box"""
    if len(img.shape) == 2:
        ## 2D
        return img[bbox[0]:(bbox[2]+1), bbox[1]:(bbox[3]+1)]
    if len(img.shape) == 3:
        #" 3D"
        if len(bbox) == 6:
            # 3D bbox
            return img[bbox[0]:(bbox[3]+1), bbox[1]:(bbox[4]+1), bbox[2]:(bbox[5]+1)]
        if len(bbox) == 4:
            # 2D bbox
            return img[:, bbox[0]:(bbox[2]+1), bbox[1]:(bbox[3]+1)]
    
def mergeBoundingBox( cell_bbox, cell_z, nuc_bbox ):
    """ Merge cell and nucleus bounding boxes """
    bbox = [0, 0, 0, 0, 0, 0]
    bbox[0] = min(cell_z, nuc_bbox[0])
    bbox[3] = max(cell_z+1, nuc_bbox[3])
    for i in range(2):
        bbox[1+i] = min(cell_bbox[i], nuc_bbox[1+i])
        bbox[3+1+i] = max(cell_bbox[2+i], nuc_bbox[3+1+i])
    return bbox

def get_boundary_cells( img, ndil=3 ):
    """ Return cells on tissu boundary in current image """ 
    dilated = binary_dilation( img > 0, disk(ndil) ) ## close small gaps, junctions
    zero = np.invert( dilated )
    zero = binary_dilation( zero, disk(ndil+2) )
    ## get pixels close to 0 and their cell value
    touching = np.unique( img[ zero ] ).tolist()
    if 0 in touching:
        touching.remove(0)
    return touching
    
def get_border_cells( img, margin=2 ):
    """ Return cells on border in current image """ 
    height = img.shape[1]
    width = img.shape[0]
    labels = list( np.unique( img[ :, 0:margin ] ) )   ## top border
    labels += list( np.unique( img[ :, (height-margin): ] ) )   ## bottom border
    labels += list( np.unique( img[ 0:margin,:] ) )   ## left border
    labels += list( np.unique( img[ (width-margin):,:] ) )   ## right border
    while 0 in labels:
        labels.remove(0)
    return labels

## make bbox and show names
def make_bbox3D(bbox_extents, mig):
    """Get the coordinates of the corners of a
    bounding box from the extents

    Parameters
    ----------
    bbox_extents : list (4xN)
        List of the extents of the bounding boxes for each of the N regions.
        Should be ordered: [min_row, min_column, max_row, max_column]

    Returns
    -------
    bbox_rect : np.ndarray
        The corners of the bounding box. Can be input directly into a
        napari Shapes layer.
    """
    minr = bbox_extents[0]*mig.scaleZ
    minc = bbox_extents[1]*mig.scaleXY
    mint = bbox_extents[2]*mig.scaleXY
    maxr = bbox_extents[3]*mig.scaleZ
    maxc = bbox_extents[4]*mig.scaleXY
    maxt = bbox_extents[5]*mig.scaleXY
    limr = (minr+maxr)/2
    
    
    bbox_rect = np.array(
        [[limr, minc, mint], [limr, minc, maxt], [limr, maxc, maxt], [limr, maxc,mint]]
    )
    bbox_rect = np.moveaxis(bbox_rect, 2, 0)

    return bbox_rect

def make_bbox2D(bbox_extents, mig):
    """Get the coordinates of the corners of a
    bounding box from the extents

    Parameters
    ----------
    bbox_extents : list (4xN)
        List of the extents of the bounding boxes for each of the N regions.
        Should be ordered: [min_row, min_column, max_row, max_column]

    Returns
    -------
    bbox_rect : np.ndarray
        The corners of the bounding box. Can be input directly into a
        napari Shapes layer.
    """
    minc = bbox_extents[0]*mig.scaleXY
    mint = bbox_extents[1]*mig.scaleXY
    maxc = bbox_extents[2]*mig.scaleXY
    maxt = bbox_extents[3]*mig.scaleXY
    
    bbox_rect = np.array(
        [[minc, mint], [minc, maxt], [maxc, maxt], [maxc,mint]]
    )
    bbox_rect = np.moveaxis(bbox_rect, 2, 0)
    return bbox_rect


def get_bblayer(lablayer, name, dim, viewer, mig):
    """ Show the cell names """
    # create the properties dictionary
    properties = regionprops_table(
        lablayer.data, properties=('label', 'bbox')
    )

    # create the bounding box rectangles
    if dim == 2:
        bbox_rects = make_bbox2D([properties[f'bbox-{i}'] for i in range(4)], mig)
    if dim == 3:
        bbox_rects = make_bbox3D([properties[f'bbox-{i}'] for i in range(6)], mig)
    if viewer.dims.ndisplay == 2:
        transl = [0,0]
    else:
        transl = [0,0,0]

    # specify the display parameters for the text
    text_parameters = {
        'text': '{label}',
        'size': 18,
        'color': 'white',
        'anchor': 'center',
        #'translation': transl,
    }

    namelayer = viewer.add_shapes(
    bbox_rects,
    face_color='transparent',
    edge_color='gray',
    edge_width = 0,
    properties=properties,
    text=text_parameters,
    name=name,
    )
    viewer.layers.select_previous()
    return namelayer


import napari
import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout
import fish_feats.MainImage as mi
import fish_feats.Configuration as cf
import fish_feats.Utils as ut
from fish_feats.NapaRNA import NapaRNA, OverlapRNA
from fish_feats.NapaCells import MainCells, Position3D, EndCells 
from fish_feats.NapaNuclei import MeasureNuclei, NucleiWidget, PreprocessNuclei 
from fish_feats.FishGrid import FishGrid
from fish_feats.NapaMix import CheckScale, CropImage, Association, Separation, CytoplasmMeasure
from fish_feats import ClassifyCells as cc
import fish_feats.FishWidgets as fwid

"""
    Handle the UI through napari plugin

    Fish&Feats proposes several actions from cell/nuclei segmentation, association, mRNA-Fish segmentation/association and quantitative measurements.

    Available under BSD License
    If you use the code or part of it, please cite associated work.

    Author: Gaëlle Letort, DSCB, Institut Pasteur/CNRS
"""

def show_documentation():
    """ Open the documentation page """
    ut.show_documentation_page("")
    return

## start without viewer for tests
def initZen():
    """ Initialize the plugin with the current viewer """
    ffeats = FishFeats()
    ffeats.init_viewer()

def startZen():
    """ Start the pipeline: open the image, get the scaling infos """
    ffeats = FishFeats()
    ffeats.init_viewer()
    return ffeats.open()

def startFromLayers():
    """ Starts the plugin on already opened image """
    ffeats = FishFeats()
    ffeats.init_viewer()
    if ffeats.viewer is None:
        ut.show_error("No viewer found")
        return
    return ffeats.openFromLayers()

def startMultiscale():
    """ Starts with multiscale view """
    ffeats = FishFeats()
    ffeats.init_viewer()
    return ffeats.startMultiscale()
    
def convert_previous_results():
    """ Convert the previous results to the new format """
    filename = ut.dialog_filename()
    if filename is None:
        print("No file selected")
        return
    ffeats = FishFeats()
    return ffeats.convert_previous_results( filename )
    

def unremove(layer):
    ut.show_info("Removing layer locked, throw an error ")
    #print(str(error))
    return

class FishFeats:
    """ Handle main interface of fishfeats """

    def __init__( self, viewer=None ):
        self.viewer = viewer
        self.cfg = None
        self.mig = None
        self.main_wid = None

        if self.viewer is None:
            self.viewer = napari.current_viewer()
            self.viewer.title = "Fish&Feats"


    #### Start
    def init_viewer( self ):
        """ Launch the plugin, initialize all """
        self.mig = mi.MainImage( talkative=True )
        self.my_cmap = ut.create_labelmap()
        self.persp = 45
        self.viewer.scale_bar.visible = True

        @self.viewer.bind_key('h', overwrite=True)
        def show_help(layer):
            ut.showHideOverlayText(self.viewer)

        @self.viewer.bind_key('F1', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 0 )

        @self.viewer.bind_key('F2', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 1 )

        @self.viewer.bind_key('F3', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 2 )

        @self.viewer.bind_key('F4', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 3 )

        @self.viewer.bind_key('F5', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 4 )

        @self.viewer.bind_key('F6', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 5 )

        @self.viewer.bind_key('F7', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 6 )

        @self.viewer.bind_key('F8', overwrite=True)
        def show_layer(viewer):
            self.show_hide( 7 )

        def show_hide( self, intlayer ):
            """ Show/hide the ith-layer """
            if 0 <= intlayer < len( self.viewer.layers ):
                self.viewer.layers[intlayer].visible = not self.viewer.layers[intlayer].visible

        @self.viewer.bind_key('Ctrl-h', overwrite=True)
        def show_shortcuts(layer):
            ut.main_shortcuts(self.viewer)

        @self.viewer.bind_key('g', overwrite=True)
        def show_grid(layer):
            self.addGrid()

        @self.viewer.bind_key('Ctrl-v', overwrite=True)
        def set_vispymode(viewer):
            pers = self.viewer.camera.perspective
            if pers > 0:
                self.persp = pers
                self.viewer.camera.perspective = 0
            else:
                self.viewer.camera.perspective = self.persp

    def open( self ):
        """ Start the pipeline: open the image, get the scaling infos """

        ## get and open the image
        filename = ut.dialog_filename()
        if filename is None:
            print("No file selected")
            return
        ut.showOverlayText(self.viewer,  "Opening image...")
        self.mig.open_image( filename=filename )
        ut.update_history(self.mig.imagedir)
        self.cfg = cf.Configuration( self.mig.save_filename(), show=False )

        ## display the different channels
        self.display_channels()
        self.endInit()
        return self.main_wid

    def openFromLayers( self ):
        """ Open from already opened layers in napari """
        ut.show_info("Loading all opened layers as channels of one image in FishFeats...")
        self.mig = mi.MainImage(talkative=True)
        if len(self.viewer.layers) == 0:
            ut.show_error("No layer(s) found")
            return None
        scale = self.viewer.layers[0].scale
        imshape = self.viewer.layers[0].data.shape
        self.mig.set_scales(scale[0], scale[1])

        ## single layer with all the channels
        if len(self.viewer.layers[0].data.shape) == 4:
            img = self.viewer.layers[0].data
            img = ut.arrange_dims( img, verbose=True )
            self.mig.set_image( img )
            ut.remove_all_layers( self.viewer )
            self.display_channels()
            return self.getImagePath()
        
        ## Or load all opened layer in the image and rename them in FishFeats style
        img = [] 
        for lay in self.viewer.layers:
            if len(lay.data.shape) == 3:
                if lay.data.shape != imshape:
                    ut.show_error("All layers should have the same shape")
                    return
                img.append( lay.data )
        self.mig.set_image(img)
        ut.remove_all_layers( self.viewer )
        self.display_channels()
        return self.getImagePath()

    def convert_previous_results(self, filename):
        """ Convert the previous results to the new format """
        self.mig = mi.MainImage( talkative=True )
        self.mig.set_imagename( filename )
    
        ## read the config file to extract the scaling and the direction
        self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
        if self.cfg.has_config():
            self.cfg.read_scale(self.mig)

        ## load cell file
        results_filename = self.mig.get_filename( endname = "_results.csv", ifexist=True )
        if results_filename != "":
            self.mig.load_from_results( results_filename )
        else:
            loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
            if loadfilename != "":
                print("Load junctions from file "+loadfilename)
                self.mig.loadCellsFromSegmentation( loadfilename )

            nucleifilename = self.mig.nuclei_filename(ifexist=True)
            if nucleifilename != "":
                print("Load nuclei from file "+str(nucleifilename))
                self.mig.load_segmentation_nuclei(nucleifilename, load_stain=False)
                self.mig.popNucleiFromMask()

        ## Load the RNA files if found some
        for chan in range(30):
            ## image was not read so todn't know how many channels are possible
            ## if find RNA file, load it
            rnafile = self.mig.rna_filename(chan=chan, how=".csv", ifexist=True)
            if rnafile == "":
                rnafile = self.mig.rna_filename(chan=chan, how=".tif", ifexist=True)
            if rnafile != "":
                self.mig.load_rnafile(rnafile, chan, topop=True)

        ## Load cytoplasmic results file if found some
        cytofile = self.mig.get_filename(endname="_cytoplasmic.csv", ifexist=True)
        if cytofile != "":
            self.mig.loadCytoplasmicTable( cytofile )
    
        ## Load features files if found some
        featfile = self.mig.get_filename(endname="_features.csv", ifexist=True)
        if featfile != "":
            self.mig.loadFeatureTable( featfile )

        ## Save the new results file
        self.mig.save_results()
        return QWidget() 

    def display_channels(self):
        """ Display the different channels of the image """
        if self.viewer is None or self.mig is None:
            return
        cmaps = ut.colormaps()
        ncmaps = len(cmaps)
        for channel in range(self.mig.nbchannels):
            cmap = cmaps[(channel%ncmaps)]
            img = self.mig.get_channel(channel)
            self.viewer.add_image( img, name="originalChannel"+str(channel), blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap=cmap, contrast_limits=ut.quantiles(img), gamma=0.9 )
        self.viewer.axes.visible = True

    def endInit(self):
        """ Finish, go to metadata step """
        self.viewer.grid.enabled = True
        ut.removeOverlayText(self.viewer)
        self.checkScale()
        return self.main_wid

    def shortcuts_window(self):
        """ Open a separate text window with the main steps and shortcuts """
        vie = napari.current_viewer()
        ut.main_shortcuts(vie)


    def getImagePath(self):
        """ Get the image path when it was open from layers """

        image_path = fwid.file_dialog( "Select image path", "All files (*)", directory=self.mig.get_image_path() )

        self.mig.set_image_path(image_path)
        ut.update_history(self.mig.imagedir)
        if self.cfg is None:
            self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
        self.endInit()
        return self.main_wid


    ### Grid tools: regular grid for spatial repere
    def addGrid(self):
        """ Interface to create/load a grid for repere """
        if "FishGrid" not in self.viewer.layers:
            grid = FishGrid(self.viewer, self.mig)
            self.viewer.window.add_dock_widget(grid, name="FishGrid")
        else:
            gridlay = self.viewer.layers["FishGrid"]
            gridlay.visible = not gridlay.visible


    def startMultiscale(self):
        """ Open the main image as multiscale for performance """
        filename = ut.dialog_filename()
        if filename is None:
            ut.show_error("No file selected, try again")
            return
        ut.showOverlayText(self.viewer,  "Opening image...")
        self.mig.open_image( filename=filename )
        ut.update_history(self.mig.imagedir)
        self.cfg = cf.Configuration(self.mig.save_filename(), show=False)
    
        for channel in range(self.mig.nbchannels):
            cmap = ut.colormapname(channel)
            img = self.mig.get_channel(channel)
            cview = self.viewer.add_image( [img, img[:,::2,::2], img[:,::4,::4] ], name="originalChannel"+str(channel), blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), colormap=cmap )
            cview.contrast_limits=ut.quantiles(img)
            cview.gamma=0.95
        self.viewer.axes.visible = True
        self.endInit()
        return self.main_wid

    def byebye( self ):
        """ Quit the pipeline """
        ut.remove_all_layers( self.viewer )
        self.viewer.title = "napari"
        ut.removeOverlayText( self.viewer )
        if self.cfg.blabla.shown():
            self.cfg.blabla.close()
        self.cfg.write_parameterfile()
        #ut.remove_widget( viewer, "Main" )
        print("Bye bye")
        #del mig
        self.main_wid.close()
        #ut.remove_all_widget( self.viewer ) 
        del self.mig
        del self.cfg

    def checkScale(self):
        """ Interface to choose the image scales and channels """
        cs = CheckScale(self)
        wid = self.viewer.window.add_dock_widget(cs, name="Scale")
        self.cfg = cs.cfg
        self.main_wid = wid

    #### Action choice
    def getChoices( self, default_action='Get cells'):
        """ Launch the interface of Main step """
        choice_wid = GetChoices( default_action, self.action_launcher )
        ut.remove_widget( self.viewer, "Main" )
        wid = self.viewer.window.add_dock_widget(choice_wid, name="Main")
        self.main_wid = wid

    def action_launcher( self, action ):
        """ Launch the specified action """
        self.cfg.addSectionText(action)
        if action == "Get cells":
            self.goJunctions()
        elif action == "Load cells from default file":
            self.loadJunctionsFile()
        elif action == "Get nuclei":
            self.getNuclei()
        elif action == "Get RNA":
            self.getRNA()
        elif action == "Get overlapping RNAs":
            self.getOverlapRNA()
        elif action == "Associate junctions and nuclei":
            self.doCellAssociation()
        elif action == "Quit plugin":
            self.byebye()
        elif action == "Image scalings":
            self.checkScale()
        elif action == "Separate junctions and nuclei":
            self.divorceJunctionsNuclei()
        elif action == "Preprocess nuclei":
            self.preprocNuclei()
        elif action == "Measure nuclear intensity":
            self.measureNuclearIntensity()
        elif action == "Measure cytoplasmic staining":
            self.cytoplasmicStaining()
        elif action == "Crop image":
            self.crop_image()
        elif action == "test":
            test()
        elif action == "Classify cells":
            if not self.mig.hasCells():
                ut.show_info("No cells - segment/load it before")
            else:
                cc.classify_cells(self.mig, self.viewer)
        elif action == "Touching labels":
            self.touching_labels()
        elif action == "3D cell positions":
            self.show3DCells()
        elif action == "Add grid":
            self.addGrid()

    def crop_image(self):
        """ Interface to crop the image and associated segmentations/results """
        crop = CropImage(self)
        self.viewer.window.add_dock_widget( crop, name="CropImage" )

    def load_all_previous_files(self):
        """ Load all the previous files with default name that it can find and init the objects accordingly """
        ## try to load separated staining
        if self.mig.should_separate():
            separated_junctionsfile = self.mig.separated_junctions_filename(ifexist=True)
            separated_nucleifile = self.mig.separated_nuclei_filename(ifexist=True)
            if (separated_junctionsfile != "") and (separated_nucleifile != ""):
                self.load_separated( separated_junctionsfile, separated_nucleifile )
    
        loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
        if loadfilename != "":
            self.cfg.addText("Load junctions from file "+loadfilename)
            self.mig.load_segmentation( loadfilename )
            self.mig.popFromJunctions()

        nucleifilename = self.mig.nuclei_filename(ifexist=True)
        if nucleifilename != "":
            self.cfg.addText("Load nuclei from file "+str(nucleifilename))
            self.mig.load_segmentation_nuclei(nucleifilename)
            self.mig.popNucleiFromMask()

        for chan in self.mig.potential_rnas():
            ## if find RNA file, load it
            rnafile = self.mig.rna_filename(chan=chan, how=".csv", ifexist=True)
            if rnafile == "":
                rnafile = self.mig.rna_filename(chan=chan, how=".tif", ifexist=True)
            if rnafile != "":
                self.mig.load_rnafile(rnafile, chan, topop=True)

        ut.remove_all_widget( self.viewer )
        self.getChoices(default_action="Classify cells")

    def load_separated( self, juncfile, nucfile ):
        """ load the separated files """
        self.mig.load_separated_staining( juncfile, nucfile )
        ut.show_info("Separated stainings loaded")

    def loadJunctionsFile(self):
        """ Load the segmentation from given file and directly init the cells """
        loadfilename = self.mig.junction_filename(dim=2,ifexist=True)
        self.cfg.addText("Load junctions from file "+loadfilename)
        self.mig.load_segmentation( loadfilename )
        self.mig.popFromJunctions()
        ut.remove_all_widget( self.viewer )
        self.getChoices(default_action="Get nuclei")

    def goJunctions(self):
        """ Choose between loading projection and cells files or recalculating """
        if self.mig.junchan is None:
            ut.show_warning( "No junction channel selected in the configuration. Go back to Image Scalings to select one." )
            return

        main_cells = MainCells( self ) 
        if main_cells.proj is not None:
            self.viewer.window.add_dock_widget( main_cells.proj, name="JunctionProjection2D" )
        else:
            self.viewer.window.add_dock_widget( main_cells, name="Get cells" )


    ################################
    ###### Show cell in 3D and possibility to edit the Z position of cells
    def show3DCells(self):
        """ Cells in 3D and update Z position of cells """
        cells_3D = Position3D( self )
        self.viewer.window.add_dock_widget(cells_3D, name="Cells in 3D")

    ##### preprocessing functions
    def preprocNuclei(self):
        """ Preprocess the nuclei before segmentation """
        preproc_nuclei = PreprocessNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(preproc_nuclei, name="Preprocess Nuclei")

    ################### Junction and nuclei separation functions

    def divorceJunctionsNuclei(self):
        """ Separate the junctions and nuclei staining if they are in the same channel """
        separation = Separation( self ) 
        self.viewer.window.add_dock_widget(separation, name="Separate")

    ################ Nuclei segmentation
    def getNuclei(self):
        """ 3D segmentation and correction of nuclei """
        print("******* 3D segmentation of nuclei ******")
        text = "Choose method and parameters to segment nuclei in 3D \n"
        text += "The nuclei are segmented from the original nuclei channel if the stainings are separate \n"
        text += "Or from the nucleiStaining image if the staining were originally mixed \n"
        ut.showOverlayText(self.viewer, text)

        ut.hide_color_layers(self.viewer, self.mig)
        ut.show_layer(self.viewer, self.mig.nucchan)
        
        nuclei_widget = NucleiWidget( self )
        self.viewer.window.add_dock_widget( nuclei_widget, name="Get nuclei" )

    ######################### Association of 2D cells with nuclei
    def doCellAssociation(self):
        """ Association of nuclei with corresponding apical junction cells """
        do_association = Association( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(do_association, name="Associating")

    ######################### RNA
    def getOverlapRNA(self):
        """ Find RNAs overlapping in several channels (non specific signal) """
        over = OverlapRNA(self.viewer, self.mig, self.cfg)
        self.viewer.window.add_dock_widget( over, name="Overlapping RNAs" )

    def getRNA(self):
        """ Segment the RNA dots in selected channels """
        if not ut.has_widget( self.viewer, "RNAs"):
            rnaGUI = NapaRNA(self.viewer, self.mig, self.cfg)


    ######## Labels edition
    def showCellsWidget(self, layerName, shapeName='CellNames', dim=3):
        """ option to show and edit cell segmentation """
        layer = self.viewer.layers[layerName]
        
        @layer.bind_key('Control-c', overwrite=True)
        def contour_increase(layer):
            if layer is not None:
                layer.contour = layer.contour + 1
        
        @layer.bind_key('Control-d', overwrite=True)
        def contour_decrease(layer):
            if layer is not None:
                if layer.contour > 0:
                    layer.contour = layer.contour - 1
        
    
        @layer.mouse_drag_callbacks.append
        def clicks_label(layer, event):
            if event.type == "mouse_press":
                if len(event.modifiers) == 0:
                    if event.button == 2:
                        label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        layer.selected_label = label
                        layer.refresh()
                    return
        
                if 'Control' in event.modifiers:
                    if event.button == 2:
                        ### Erase a label
                        label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        layer.data[layer.data==label] = 0
                        layer.refresh()
                        return

                    if event.button == 1:
                        ## Merge two labels
                        start_label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        yield
                        while event.type == 'mouse_move':
                            yield
                        end_label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        # Control left-click: merge labels at each end of the click
                        print("Merge label "+str(start_label)+" with "+str(end_label))
                        frame = None
                        if layer.ndim == 3:
                            frame = int(self.viewer.dims.current_step[0])
                        ut.merge_labels( layer, frame, start_label, end_label )
    

        @layer.bind_key('m', overwrite=True)
        def set_maxlabel(layer):
            mess = "Max label on image "+shapeName+": "+str(np.max(layer.data)) 
            mess += "\n "
            mess += "Number of labels used: "+str(len(np.unique(layer.data)))
            ut.show_info( mess )
            layer.mode = "PAINT"
            layer.selected_label = np.max(layer.data)+1
            if layer.selected_label == 1:
                layer.selected_label = 2
            layer.refresh()
            return

        @layer.bind_key('l', overwrite=True)
        def switch_show_lab(layer):
            if shapeName in self.viewer.layers:
                self.viewer.layers.remove(shapeName)
            else:
                ut.get_bblayer(layer, shapeName, dim, self.viewer, self.mig)
            return
    
        def relabel_layer():
            maxlab = np.max(layer.data)
            used = np.unique(layer.data)
            nlabs = len(used)
            if nlabs == maxlab:
                print("already relabelled")
                return
            for j in range(2, nlabs+1):
                if j not in used:
                    layer.data[layer.data==maxlab] = j
                    maxlab = np.max(layer.data)
            layer.refresh()

        def show_names():
            if shapeName in self.viewer.layers:
                self.viewer.layers.remove(shapeName)

        help_text = ut.labels_shortcuts( level = 0 )
        header = ut.helpHeader(self.viewer, layerName)
        ut.showOverlayText(self.viewer, header+help_text)
    
        print( "\n #########################################\n Labels correction options:\n " + help_text + self.textCellsWidget() )

        if "Junctions" in self.viewer.layers:
            self.viewer.layers["Junctions"].preserve_labels = True

    def textCellsWidget(self):
        text = "  <Control+left click> from one label to another to merge them (the label kept will be the last one) \n"
        text += "'show_cellnames' (<l>) add a new layer showing the label (number) around each object position. \n"
        #text += "'relabel update' the cell labels to have consecutives numbers from 2 to number_of_cells.\n"
        text += "\n For 3D: \n"
        text += "In 3D, most label actions wont work if Vispy perspective is ON. Switch it off with 'Ctrl-v' before.\n"
        text += "If n_edit_dim is set on 3 (top left panel), edition will affect all or several z (slices) \n"
        text += "If n_edit_dim is set on 2, edition will only affect the active slice \n"
        return text
    
    def correction_junctions( self ):
        """ Manual correction of segmentation step """
        ut.removeOverlayText( self.viewer )
        maskview = self.viewer.add_labels( self.mig.junmask, blending='additive', scale=(self.mig.scaleXY, self.mig.scaleXY), name="Junctions" )
        maskview.contour = 3
        maskview.selected_label = 2
        self.showCellsWidget( "Junctions", shapeName="JunctionsName", dim=2 )
        #ut.remove_widget( self.viewer, "Get cells" )
        # saving and finishing
        endcells = EndCells( self )
        self.viewer.window.add_dock_widget( endcells, name="EndCells" )


    def measureNuclearIntensity(self):
        """ Measure intensity inside segmented nuclei """
        if not self.mig.hasNuclei():
            ut.show_warning( "Segment/Load nuclei before" )
            return
        meas_nuc = MeasureNuclei( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(meas_nuc, name="Measure nuclei")

    ############ measure cyto
    def cytoplasmicStaining(self):
        """ Measure the cytoplasmic signal close to the apical surface """
        cytoMeas = CytoplasmMeasure( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(cytoMeas, name="Measure cytos")

    def helpMessageEditContours(self, dim=2):
        text = '- To see the cells as filled area, put 0 in the *contour* field. Else to see the contours, in *contour* field, put 1 or more (will be thicker if >1) \n'
        text = text + '- To erase one label entirely, put *0* in the *label* field, select the *fill* tool and click on the label \n'
        text = text + '- To add one label, choose a label value higher than all the ones in the image,'
        text = text + ' put it in the *label* field, select the *drawing* tool and draw it.'
        text = text + ' Fill the contour you have drawn to finish the new cell \n'
        text = text + '- To draw, choose the label value in the *label* field, and click and drag to paint.'
        text += ' If *preserve labels* is selected, drawing above another label doesn t affect it \n'
        text += '- Holding space to zoom/unzoom \n'
        #text += 'You can set the selected label to be one larger than the current largest label by pressing M.\n'
        if dim == 3:
            text += '- Check n_edit_dimensions box to modify all 3D nuclei at once (else work only on the current slice) \n'
        return text

############################## Extra-tools

    ### Touching labels for Griottes
    def touching_labels(self):
        """ Dilate labels so that they all touch """
        ## perform the label expansion
        from skimage.morphology import binary_opening
        from skimage.segmentation import expand_labels
        print("********** Generate touching labels image ***********")
    
        ## get junctions img
        if "Cells" not in self.viewer.layers:
            if self.mig.pop is None or self.mig.pop.imgcell is None:
                ut.show_info("Load segmentation before!")
                return
            labimg = self.viewer.add_labels(self.mig.pop.imgcell, name="Cells", scale=(self.mig.scaleXY, self.mig.scaleXY), opacity=1, blending="additive")
            labimg.contour = 0

        ## skeletonize it
        img = self.viewer.layers["Cells"].data
        ext = np.zeros(img.shape, dtype="uint8")
        ext[img==0] = 1
        ext = binary_opening(ext, footprint=np.ones((2,2)))
        newimg = expand_labels(img, distance=4)
        newimg[ext>0] = 0
        newlay = self.viewer.add_labels(newimg, name="TouchingCells", scale=(self.mig.scaleXY, self.mig.scaleXY), opacity=1, blending="additive")
        newlay.contour = 0

        ## open the widget for options
        tlabel_wid = TouchingLabels( self.viewer, self.mig, self.cfg )
        self.viewer.window.add_dock_widget(tlabel_wid, name="Touching labels")

class TouchingLabels( QWidget):
    """ Generate an image with touching labels from the junctions image, handle compability with Griottes """
    def __init__(self, viewer, mig, cfg):
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg

        layout = QVBoxLayout()
        ## save the image
        save_btn = fwid.add_button( "Save touching labels image", self.save_touching_labels_image, descr="Save the resulting image of expanded cell labels", color=ut.get_color("save") )  
        layout.addWidget( save_btn )
        
        ## if Griottes has run, adapt it to the scale
        scale_btn = fwid.add_button( "Scale Griottes image", self.scale_griottes, descr="Scale the images resulting from Griottes computing to the main image scale" )  
        layout.addWidget( scale_btn )
        self.setLayout(layout)
    
    def save_touching_labels_image( self ):
        """ Save the touching labels image """
        if "TouchingCells" not in self.viewer.layers:
            ut.show_error("No touching labels image to save")
            return
        outname = self.mig.build_filename( "_touching_labels.tif")
        self.mig.save_image(self.viewer.layers["TouchingCells"].data, imagename=outname)
        ut.show_info("Saved touching labels image as "+outname)

    def scale_griottes( self ):
        """ Scale the Griottes images to the main image scale """
        ut.scale_layer( self.viewer, "Centers", (self.mig.scaleXY, self.mig.scaleXY) )
        ut.scale_layer( self.viewer, "Contact graph", (self.mig.scaleXY, self.mig.scaleXY) )
        ut.scale_layer( self.viewer, "Graph", (self.mig.scaleXY, self.mig.scaleXY) )


def test():
    print("for test")
    #img = mig.tryDiffusion( nchan=2 )
    #viewer.add_image(img, name="test", scale=(mig.scaleZ, mig.scaleXY, mig.scaleXY), colormap=my_cmap, blending="additive")


class GetChoices( QWidget ):
    """ Main widget with all the action choices """

    def __init__( self, default_action, action_fnc ):
        """ Initialiaze the Main interface with the choice of action to perform """
        super().__init__()
        self.default_action = default_action
        self.action_fnc = action_fnc

        layout = QVBoxLayout()
        ## Choice list
        action_line, self.action = fwid.list_line( "Action: ", descr="Choose which action (step) to perform now", func=None )
        choices = ['Get cells', 'Get nuclei', 'Associate junctions and nuclei', 'Get RNA', 'Get overlapping RNAs', 'Measure cytoplasmic staining', 'Measure nuclear intensity', 'Image scalings', 'Separate junctions and nuclei', 'Preprocess nuclei', '3D cell positions', 'Quit plugin', 'Classify cells', 'Touching labels', 'Add grid', 'Crop image']
        for choice in choices:
            self.action.addItem(choice)
        self.action.setCurrentText( self.default_action )
        layout.addLayout( action_line )
        ## button go
        go_btn = fwid.add_button( "GO", self.launch_action, descr="Launch the selected action", color=ut.get_color("go") )
        layout.addWidget(go_btn)
        self.setLayout(layout)
        self.action.currentTextChanged.connect( self.launch_action )

    def launch_action(self):
        """ Launch next step with selected action """
        action = self.action.currentText()
        self.action_fnc( action )




import os
import napari
from magicgui import magicgui, magic_factory
import fish_feats.MainImage as mi
import fish_feats.Configuration as cf
from napari.utils.translations import trans

def get_directory(title=""):
    try:
        from tkinter import Tk
        from tkFileDialog import askopenfilenames
    except:
        from tkinter import Tk
        from tkinter import filedialog

    Tk().withdraw() # we don't want a full GUI, so keep the root window from appearing
    filenames = filedialog.askdirectory(title=title) # show an "Open" dialog box and return the path to the selected file
    return filenames

def get_filename(mdir, title=""):
    try:
        from tkinter import Tk
        from tkFileDialog import askopenfilenames
    except:
        from tkinter import Tk
        from tkinter import filedialog

    Tk().withdraw() # we don't want a full GUI, so keep the root window from appearing
    filenames = filedialog.askopenfilenames(initialdir=mdir, title=title) # show an "Open" dialog box and return the path to the selected file
    return filenames[0]

def start():
    global viewer
    viewer = napari.current_viewer()
    viewer.title = "ZENnapari"
    thedirname = get_directory("Select folder to process")
    print(thedirname)

    configname = get_filename(mdir=thedirname, title="Select config file (.fishfeats)")
    cfg = cf.Configuration(configname, show=False)
    mig = mi.MainImage( talkative=True )
    listfiles = os.listdir(thedirname)
    
    dofiles = []
    for cfile in listfiles:
        if cfile.endswith(".tif") or cfile.endswith(".czi") or cfile.endswith(".ims"):
            dofiles.append(cfile)

    return check_files(dofiles, mig, cfg, thedirname)

def check_files(filelist, mig, cfg, mpath):
     
    @magicgui(call_button = "Go",
            select_files=dict(widget_type="Select", choices=filelist))
    def get_files( select_files ):
        for cfile in select_files:
            showOverlayText( "Doing file "+cfile, size=14 )
            fullfile = os.path.join(mpath, cfile)
            go_one_file( fullfile, mig, cfg)

    widg = viewer.window.add_dock_widget(get_files, name="Choose files")
    return widg

def go_one_file(fullfile, mig, cfg):
    mig.open_image(filename=fullfile)
    cfg.read_scale(mig)
    paras = cfg.param_junctions()
    roijunc = mig.prepare_segmentation_junctions()
    mig.do_segmentation_junctions( paras["Junc_method"], roijunc )
    filename = mig.junction_filename(dim=2, ifexist=False)
    mig.save_image( mig.junmask, filename, hasZ=False )
    print("Segmented file "+filename+" saved")

def showOverlayText( text, size=12 ):
    viewer.text_overlay.text = trans._(text)
    viewer.text_overlay.visible = True
    viewer.text_overlay.font_size = size



import numpy as np
import os
import fish_feats.MainImage as mi
import napari

def starting():
    """ test initialization of the napari plugin """
    viewer = napari.Viewer( show=False )
    from fish_feats.Naparing import init_viewer
    init_viewer(viewer)


mig = mi.MainImage(talkative=True)
    
def test_load_image():
    test_img = "./_tests/files/imaris0.ims"
    mig.open_image(test_img)
    assert mig.image is not None
    assert mig.nbchannels == 4 
    assert mig.free_channel() == 1
    assert mig.build_filename("_end.tif") == os.path.join( ".", "_tests", "files", "results", "imaris0_end.tif" )

def read_config():
    """ read the configuration file """
    import fish_feats.Configuration as cf
    cfg = cf.Configuration( mig.save_filename(), show=False )
    ## check that scale was read from the file metatdata
    assert mig.scaleZ == 0.19
    if cfg.has_config():
        cfg.read_scale( mig )
        assert mig.zdirection == -1 
        ## check taht the scale was read from the config file
        assert mig.scaleXY == 0.166
        assert mig.scaleZ == 0.2

        #" Cehck reading a parameter set from the config file"
        paras = cfg.read_parameter_set("JunctionSeg")
        assert paras is not None
        assert "method" in paras


def test_junctions_separation():
    assert mig.should_separate() == True
    mig.separate_junctions_nuclei()
    assert mig.nucstain is not None
    assert mig.should_separate() == False

def test_load_segfile():
    """ test loading from the segmentation file """
    assert mig.hasCells() == False
    segfile = os.path.join( ".", "_tests", "files", "results", "imaris0_cells2D.tif" )
    mig.load_segmentation( segfile )
    mig.popFromJunctions()
    assert mig.hasCells() == True
    assert mig.nbCells() == 58

def test_load_nucleifile():
    """ Test loading nuclei segmentation """
    nucfile = os.path.join( ".", "_tests", "files", "results", "imaris0_nuclei.tif" )
    mig.load_segmentation_nuclei(nucfile)
    assert mig.nucstain is not None
    assert mig.pop.imgnuc is not None

def test_nuclei_filter():
    mig.popNucleiFromMask(associate=False)
    assert mig.hasNuclei() == True 
    assert mig.nbNuclei() == 48
    mig.pop.relabelNuclei()
    assert mig.nbNuclei() == 48
    mig.filterNuclei( 100 )
    assert mig.nbNuclei() == 39

def test_load_rnafile():
    rnafile = os.path.join( ".", "_tests", "files", "results", "imaris0_RNA1.csv" )
    mig.load_rnafile( rnafile, 1 )
    assert mig.rnas[1].nspots() >= 1200

def test_association():
    mig.go_association( distance = 10 )
    assert len(mig.pop.association.keys()) == 20

def test_rna_segmentation():
    mig.find_rna( 2, 1000, 250, True, None )
    assert mig.rnas[2].nspots() >= 1000
    assert mig.rnas[2].nspots() <= 2500
    assert mig.get_rna_threshold(2) >= 3.5 
    assert mig.get_rna_threshold(2) <= 4.5 

def test_rna_assignement():
    mig.assign_rna(2, "Projection", 10, 0)
    assert len(np.unique(mig.rnas[2].labels)) >= 40

if __name__ == "__main__":
    starting()
    test_load_image()
    read_config()
    test_junctions_separation()
    test_load_nucleifile()
    test_load_segfile()
    test_nuclei_filter()
    test_association()
    test_rna_segmentation()
    test_rna_assignement()
    test_load_rnafile()
    print("Everything passed")



import numpy as np
import cv2
import scipy.ndimage as ndimage
from napari.utils.notifications import show_info
from math import floor
import time
from napari.utils import progress

def sepanet( img, sepdir, patchsize=256 ):
    """ Separate junctions and nuclei with trained DL """
    print("sepaNet with models in "+str(sepdir))

    ## to allocate more gpus, try
    import tensorflow as tf
    print(tf.test.is_built_with_cuda())
    print(tf.__version__)
    print("Num GPUs Available: ", len(tf.config.list_physical_devices('GPU')))
    #config = tf.compat.v1.ConfigProto(
      #device_count = {'GPU': 0}
    #)
    #sess = tf.compat.v1.Session(config=config)
    #tf.compat.v1.keras.backend.set_session(sess)

    # load model
    model_path = sepdir
    model = tf.keras.models.load_model(model_path, custom_objects={"mse_two":mse_two, "both_MSE":both_MSE, "both_MSE_percent_0":both_MSE_percent_0, "both_MSE_percent_1":both_MSE_percent_1})
    res = run_on_image( img, model, patchsize, step=50 )
    return res[:,:,:,0], res[:,:,:,1]

def run_on_image(imgtest, model, patchsize, step=50):
    imgtest.astype(float)
    imgtest = normalise(imgtest)
    imgtest = smooth(imgtest)

    ## handle case of very small image, smaller than the patchsize
    smallimg = None
    if (imgtest.shape[1] < patchsize) or (imgtest.shape[2] < patchsize):
        smallimg = np.copy(imgtest)
        imgtest = np.zeros((imgtest.shape[0], patchsize, patchsize)) 
        imgtest[:,:smallimg.shape[1], :smallimg.shape[2] ] = smallimg

    ## split in patches
    imshape = imgtest.shape
    shapey = imshape[1]
    shapex = imshape[2]
    resimg = np.zeros(imshape+(2,), dtype="float")

    #for z, zimg in enumerate(imgtest):
    for z in progress(range(imgtest.shape[0])):
        zimg = imgtest[z]
        patchs = []
        nimg = np.zeros(imshape[1:3]+(2,), dtype="uint8")
        inds = []
        for y in range(floor(shapey/step)):
            ey = min(y*step+patchsize, shapey)
            sy = ey - patchsize
            for x in range(floor(shapex/step)):
                ex = x*step+patchsize
                ex = min(ex, shapex)
                sx = ex - patchsize
                patch = zimg[sy:ey, sx:ex]
                nimg[sy:ey, sx:ex,0] = nimg[sy:ey, sx:ex,0] + 1
                nimg[sy:ey, sx:ex,1] = nimg[sy:ey, sx:ex,1] + 1
                patchs.append(patch)
                inds.append((sy, ey, sx, ex))
        patchs = np.array(patchs)
        res = model.predict(patchs)
        for ind, (sy, ey, sx, ex) in enumerate(inds):
            resimg[z,sy:ey, sx:ex,:] += res[ind]/nimg[sy:ey,sx:ex]

    ## get back the original image if it was too small
    if smallimg is not None:
        resimg = resimg[:, :smallimg.shape[1], :smallimg.shape[2]]
    resimg = np.uint8(resimg*255)
    return resimg


def normalise(img):
    quants = np.quantile(img, [0.1, 0.99])
    img = (img - quants[0]) / (quants[1]-quants[0])
    img = np.clip(img, 0, 1)
    return img

def smooth(img):
    for z in range(img.shape[0]):
        img[z,] = ndimage.gaussian_filter(img[z,], 1)
    return img

def both_MSE( y_true, y_pred ):
    acc0 = keras.metrics.mean_squared_error(y_true[:,:,:,0], y_pred[:,:,:,0])
    acc1 = keras.metrics.mean_squared_error(y_true[:,:,:,1], y_pred[:,:,:,1])
    return acc0 + acc1

def both_MSE_percent_0( y_true, y_pred ):
    acc0 = keras.metrics.mean_absolute_percentage_error(y_true[:,:,:,0], y_pred[:,:,:,0])
    return acc0
def both_MSE_percent_1( y_true, y_pred ):
    acc0 = keras.metrics.mean_absolute_percentage_error(y_true[:,:,:,1], y_pred[:,:,:,1])
    return acc0


def mse_two(y_true, y_pred):
    y_true = tf.reshape(y_true, [-1])
    y_pred = tf.reshape(y_pred, [-1])
    mse = tf.keras.losses.MeanSquaredError()
    return mse(y_true, y_pred)

### Separation based on filterings
def junctionsCoherence(img, medblur=3, quant=0.98, dsig=3, cornersig=5, ratio=0.5, niter=4):
    ## Coherence enhancing diffusion, Weickert et al.
    #from skimage import exposure
    height, width = img.shape[:2]
    qmax = np.quantile(img, quant)
    qmin = np.min(img)
    img = np.uint8( (img-qmax)/(qmax-qmin)*255 )
    #img = cv2.normalize(src=img, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    img = cv2.medianBlur(img,medblur)
    #img = exposure.adjust_gamma(img, 0.7)

    for i in range(niter):
        ## get the features: eigen values
        cur = cv2.normalize(src=img, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
        eigenV = cv2.cornerEigenValsAndVecs(cur, cornersig, 3)  ## block size, eigen val and vec to edge/croner detection
        eigenV = eigenV.reshape(height, width, 3, 2)  # [[e1, e2], v1, v2]
        x1, y1 = eigenV[:,:,1,0], eigenV[:,:,1,1]

        ## derivatives
        gxx = cv2.Sobel(cur, cv2.CV_32F, 2, 0, ksize=dsig)
        gxy = cv2.Sobel(cur, cv2.CV_32F, 1, 1, ksize=dsig)
        gyy = cv2.Sobel(cur, cv2.CV_32F, 0, 2, ksize=dsig)
        gvv = x1*x1*gxx + 2*x1*y1*gxy + y1*y1*gyy
        m = gvv < 0
        
        ## combine results
        eroded = cv2.erode(img, None)
        dilated = cv2.dilate(img, None)
        imgt = eroded
        imgt[m] = dilated[m]
        img = (img*(1.0 - ratio) + imgt*ratio)
    return img

def anisoDiff(img, iterations=1):
    from medpy.filter.smoothing import anisotropic_diffusion
    return anisotropic_diffusion(img, niter=iterations)

def topHat(img, xyrad):
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (xyrad,xyrad))
    resimg = np.copy(img)
    i = 0
    for zimg in img:
        resimg[i] = cv2.morphologyEx(zimg, cv2.MORPH_TOPHAT, kernel)
        i = i + 1
    return resimg

def topHat2D(img, xyrad):
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (xyrad,xyrad))
    return cv2.morphologyEx(img, cv2.MORPH_TOPHAT, kernel)

def removeOutliers(img, rz=2, rxy=4, threshold=50):
    ## find outliers
    nimg = cv2.normalize(src=img, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    med = ndimage.median_filter(nimg, size=(rz,rxy,rxy) )
    outliers = nimg>(med+threshold)
    ## replace outliers by neighboring median value
    imedian = ndimage.median_filter(img, size=(rz,rxy,rxy) )
    img[outliers] = imedian[outliers]
    return img

def removeOutliersIn2D(img, rxy=3, threshold=50):
    ## find outliers
    resimg = np.copy(img)
    for z in range(img.shape[0]):
        zimg = img[z,]
        nimg = cv2.normalize(src=zimg, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
        med = ndimage.median_filter(nimg, size=(rxy,rxy) )
        outliers = nimg>(med+threshold)
        ## replace outliers by neighboring median value
        imedian = ndimage.median_filter(zimg, size=(rxy,rxy) )
        resimg[z][outliers] = imedian[outliers]
    return resimg


def separateNucleiJunc(img, outrz=1, outrxy=6, threshold=40, edge=5, space=100, zhatrad=1, hatrad=4):
    show_info("Discriminating between nuclei and junction staining...")
    # remove background
    img = cv2.normalize(src=img, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    img = removeOutliers(img, rz=outrz, rxy=outrxy, threshold=threshold)

    ## smooth - edge preserving
    imgjun = np.copy(img)
    for im in range(imgjun.shape[0]):
        imgjun[im,] = cv2.bilateralFilter(imgjun[im,], edge, space, space)

    # get junctions: small structures but not too small (dots)
    imgjun = ndimage.white_tophat(input=imgjun, size=(zhatrad,hatrad,hatrad))
    imgdots = ndimage.white_tophat(input=imgjun, size=(0,2,2))
    imgjun = imgjun-1.5*imgdots
    imgjun[imgjun<0] = 0

    # get nuclei   
    #imgmin = ndimage.minimum_filter(img, size=(2,25,25))
    #img = img - imgmin
    #img = ndimage.median_filter(img, size=(zhatrad*1,hatrad*1,hatrad*1) )
    #imgblurjunc = cv2.normalize(src=imgjun, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    imgblurjunc = ndimage.uniform_filter(imgjun, size=(int(zhatrad*2),int(hatrad*1),int(hatrad*1)) )
    #imgsmall = ndimage.white_tophat(input=img, size=(zhatrad,int(hatrad*2),int(hatrad*2)) )
    img = img - imgblurjunc*8
    img[img<0]=0
    img = ndimage.uniform_filter(img, size=(zhatrad*1,hatrad*1,hatrad*1) )
    print("Done")
    return img, imgjun

def separateNucleiJuncV0(img, rz, rxy, zhatrad, hatrad):
    print("Discriminating between nuclei and junction staining...")
    # remove background
    imgmin = ndimage.minimum_filter(img, size=(2,30,30))
    img = img - imgmin
    img = removeOutliers(img, rz=rz, rxy=rxy, threshold=100)
    # smooth image
    #img = ndimage.uniform_filter(img, size=(rz,rxy,rxy) )
    img = ndimage.median_filter(img, size=(rz,rxy,rxy) )
    #img = anisoDiff(img, iterations=5)
    #img = ndimage.uniform_filter(img, size=(rz,rxy,rxy) )
    #img = ndimage.gaussian_filter(img, sigma=(0.5, 1, 1))
    # Top-hat filter to separate small and big structures
    ##kernel = ndimage.generate_binary_structure(rank=3, connectivity=3)
    #imgjun = ndimage.white_tophat(input=img, size=(int(zhatrad/2),int(hatrad/2),int(hatrad/2)))
    imgjun = ndimage.white_tophat(input=img, size=(zhatrad,hatrad,hatrad))
    imgsmall = ndimage.white_tophat(input=img, size=(zhatrad,int(hatrad*1.5),int(hatrad*1.5)) )
    #img = exposure.adjust_gamma(img, 0.9)
    imgsmall = ndimage.gaussian_filter(imgsmall, sigma=(0.7, 1.2, 1.2))
    img = img - imgsmall*1.4
    img[img<0]=0
    imgjun = ndimage.gaussian_filter(imgjun, sigma=(0.7, 1.2, 1.2))
    #img = ndimage.gaussian_filter(img, sigma=(0.5, 1.5, 1.5))
    print("Done")
    return img, imgjun

def smoothNuclei(img, radxy, radz):
    return ndimage.uniform_filter(img, size=(radz,radxy,radxy) )


